# coding: utf-8
"""
Objects related to the computation of Debye-Waller tensors from the generalized phonon DOS.
"""
from __future__ import annotations

import numpy as np
import pandas as pd
import abipy.core.abinit_units as abu

from collections import OrderedDict
from monty.string import list_strings, marquee
from monty.collections import dict2namedtuple
from monty.termcolor import cprint
from abipy.core.structure import Structure
from abipy.core.mixins import Has_Structure
from abipy.tools.plotting import add_fig_kwargs, set_axlims, get_axarray_fig_plt, set_visible
from abipy.tools.printing import print_dataframe


class _Component:
    """
    Object used to select/plot the components of the DW tensor.
    """

    def __init__(self, name, ij, **plot_kwargs):
        self.name = name
        self.ij = ij
        self.i, self.j = None, None
        if self.ij is not None:
            self.i, self.j = self.ij[0], self.ij[1]
        self.plot_kwargs = plot_kwargs

    def get_tavg_label(self, what, with_units=False):
        n = dict(displ="U", vel="V")[what]
        unit = ""
        if with_units:
            unit = r"\;%s" % (dict(displ=r"\AA^2", vel="v")[what])

        if self.name == "trace":
            return r"$\langle {%s}^2 \rangle%s$" % (n, unit)
        else:
            return r"$\langle {%s}_{%s} \rangle%s$" % (n, self.name, unit)

    def eval33w(self, mat33w):
        #assert mat33w.shape[:2] == (3, 3)
        if self.ij is not None:
            return mat33w[self.i, self.j]
        if self.name == "trace":
            return mat33w.trace()

        raise TypeError("Don't know how to extract data for `%s`" % self.name)


class MsqDos(Has_Structure):
    """
    This object stores the generalized phonon DOS in CARTESIAN coords.
    This DOS-like quantity allows one to calculate Debye Waller factors as a function of Temperature
    by integration with 1/omega and the Bose-Einstein factor.
    This object is usually instanciated via the read_msq_dos method of the PhdosReader
    and stored as attribute of the PhdosFile instance.

    See :cite:`Lee1995` for the further details about the internal implementation and
    :cite:`Trueblood1996` for the different conventions used by crystallographers.

    See also http://atztogo.github.io/phonopy/formulation.html#thermal-displacement
    """
    C = _Component

    ALL_COMPS = OrderedDict([
        ("trace", C(name="trace", ij=None, color="k")),
        ("xx", C(name="xx", ij=(0, 0), color="r", ls="-")),
        ("yy", C(name="yy", ij=(1, 1), color="g", ls="-")),
        ("zz", C(name="zz", ij=(2, 2), color="b", ls="-")),
        ("xy", C(name="xy", ij=(0, 1), color="c", ls="--")),
        ("xz", C(name="xz", ij=(0, 2), color="m", ls="--")),
        ("yz", C(name="yz", ij=(1, 2), color="y", ls="--")),
        # Symmetric components.
        #("yx", (1, 0)),
        #("zx", (2, 0)),
        #("zy", (2, 1)),
    ])
    del C

    def __init__(self, structure, wmesh, values, amu_symbol):
        """
        Arg:
            structure: |Structure| object.
            wmesh: Frequency mesh in eV
            values: (natom, 3, 3, nomega) array with generalized DOS in Cartesian coordinates.
            amu_symbol: Dictionary element.symbol -> mass in atomic units.
        """
        self._structure = structure
        self.wmesh = wmesh * abu.eV_Ha ####
        self.nw = len(self.wmesh)
        self.values = values * abu.Ha_eV ###
        self.amu_symbol = amu_symbol
        assert len(self.values) == len(self.structure)

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self._structure

    def __str__(self) -> str:
        """Invoked by str"""
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """
        Human-readable string with useful information on the object.

        Args:
            verbose: Verbosity level.
        """
        lines = []; app = lines.append

        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(marquee(r"Fullfilment of \int dw g_ij(w) = \delta_ij", mark="="))
        app("")
        from scipy.integrate import simps
        for iatom, site in enumerate(self.structure):
            d = simps(self.values[iatom], x=self.wmesh)
            app("For site: %s" % site)
            app(str(d))
            app("Trace: %.4f, determinant: %.4f" % (d.trace(), np.linalg.det(d)))
            app("")

        for fmt in ("cartesian", "cif"):
            df = self.get_dataframe(temp=300, view="inequivalent", fmt=fmt, verbose=verbose)
            s = print_dataframe(df, title="Format: %s" % fmt, file="string")
            lines.extend(s.splitlines())

        title = marquee("Constraints on tensor components in reduced coords induced by site symmetries", mark="=")
        s = print_dataframe(self.structure.site_symmetries.get_tensor_rank2_dataframe(), file="string", title=title)
        lines.extend(s.splitlines())

        #if verbose > 1:
            #max_err = self.check_site_symmetries(verbose=verbose)

        return "\n".join(lines)

    def get_json_doc(self, tstart=0, tstop=600, num=11) -> dict:
        """
        Return dictionary with results. Used by emmet builder.

        Args:
            tstart: The starting value (in Kelvin) of the temperature mesh.
            tstop: The end value (in Kelvin) of the mesh.
            num: int, optional Number of samples to generate.
        """
        tmesh = np.linspace(tstart, tstop, num=num)
        # (natom, 3, 3, nt)
        ucart_t = self.get_msq_tmesh(tmesh, what_list=("displ")).displ

        # Convert tensor to U_CIF format
        ucif_t = np.empty_like(ucart_t)
        for it in range(num):
            ucif_t[:,:,:,it] = self.convert_ucart(ucart_t[:,:,:,it], fmt="cif")

        jdoc = {
            "natom": len(self.structure),
            "nomega": self.nw,             # Number of frequencies
            "ntemp": len(tmesh),            # Number of temperatures
            "tmesh": tmesh,                 # Temperature mesh in K
            "wmesh": self.wmesh,            # Frequency mesh in ??
            "gdos_aijw": self.values,       # Generalized DOS in Cartesian coords. (natom, 3, 3, nomega) array.
            "amu_symbol": self.amu_symbol,  # Dict symbol --> Atomic mass in a.u.
            "structure": self.structure,    # Structure object
            "ucif_t": ucif_t,               # U tensors (natom, 3, 3, ntemp)  as a function of T for T in tmesh in CIF format.
            "ucif_string_t300k": self.get_cif_string(temp=300, symprec=None),  # String with U tensor at T=300 in Cif format.
        }

        return jdoc

    def get_msq_tmesh(self, tmesh, iatom_list=None, what_list=("displ", "vel")):
        """
        Compute mean square displacement for each atom in `iatom_list` as a function of T.
        in Cartesian coordinates and atomic-units.

        Args:
            tmesh: array-like with temperatures in Kelvin.
            iatom_list: List of atom sites to compute. None if all aomts are wanted.
            what_list: "displ" for displacement, "vel" for velocity tensor.

        Return:
            namedtuple with (tmesh=tmesh, displ=msq_d, vel=msq_v)

            msq_d = np.empty((natom, 3, 3, nt))
        """
        tmesh = np.array(tmesh)
        nt = len(tmesh)

        # Frequency mesh starts at iomin to avoid 1/0 and ignore eventual negative frequencies.
        for iomin, w in enumerate(self.wmesh):
            if w > 1e-12: break
        else:
            raise ValueError("Cannot find index such that wmesh[i] > 1e-12 !!!")
        wvals = self.wmesh[iomin:]
        nw = len(wvals)

        # We will compute: Ucart(T, k, ij) = 1/M_k \int dw (n(w) + 1/2) g_ij(w) / w for the k-atom in a.u.
        # Calculate Bose-Einstein occupation factors only once for each T (instead of for each atom).
        npht = np.zeros((nt, nw))
        for it, temp in enumerate(tmesh):
            npht[it] = abu.occ_be(wvals, temp * abu.kb_HaK) + 0.5

        natom = len(self.structure)
        msq_d = np.empty((natom, 3, 3, nt))
        msq_v = np.empty((natom, 3, 3, nt))
        what_list = list_strings(what_list)

        # Perform frequency integration to get tensor(T)
        from scipy.integrate import simps
        if iatom_list is not None: iatom_list = set(iatom_list)
        for iatom in range(natom):
            if iatom_list is not None and iatom not in iatom_list: continue
            symbol = self.structure[iatom].specie.symbol
            for it in range(nt):
                fn = self.values[iatom, :, :, iomin:] * npht[it]
                if "displ" in what_list:
                    # Mean square displacement for each atom as a function of T (bohr^2).
                    ys = fn / wvals
                    fact = 1.0 / (self.amu_symbol[symbol] * abu.amu_emass)
                    msq_d[iatom, :, :, it] = simps(ys, x=wvals) * fact * abu.Bohr_Ang ** 2
                if "vel" in what_list:
                    # Mean square velocity for each atom as a function of T (bohr^2/atomic time unit^2)"
                    ys = fn * wvals
                    fact = 1.0 / (self.amu_symbol[symbol] * abu.amu_emass)
                    msq_v[iatom, :, :, it] = simps(ys, x=wvals) * fact # * abu.velocity_at_to_si ** 2

        return dict2namedtuple(tmesh=tmesh, displ=msq_d, vel=msq_v)

    def convert_ucart(self, ucart_mat, fmt):
        """
        Convert the U tensor from Cartesian coordinates to format `fmt`
        Return new tensor. See also :cite:`Grosse-Kunstleve2002`.

        Args:
            ucart_mat: (natom,3,3) array with tensor in Cartesian coords.
            fmt: Output format. Available options: "cif", "ustar", "beta", "cartesian"

        Return: (natom, 3, 3) tensor.
        """
        natom = len(self.structure)
        if fmt == "cartesian":
            return ucart_mat.copy()

        #elif fmt == "B":
        #    # Eq 7
        #    return ucart_mat * 8 * np.pi**2

        elif fmt in ("cif", "ustar", "beta"):
            # Build A matrix
            amat = self.structure.lattice.matrix.T
            ainv = np.linalg.inv(amat)
            new_mat = np.zeros_like(ucart_mat)
            # Eq 3b: A^-1 U_Cart A^-T
            for iatom in range(natom):
                new_mat[iatom] = np.matmul(ainv, np.matmul(ucart_mat[iatom], ainv.T))

            # Now we have Ustar
            if fmt == "ustar": return new_mat
            if fmt == "beta": return new_mat * 8 * np.pi**2  # Eq 6

            # CIF Format Eq 4a
            # Build N matrix (no 2 pi factor)
            lengths = self.structure.lattice.reciprocal_lattice_crystallographic.lengths
            ninv = np.diag(1.0 / np.array(lengths, dtype=float))
            # N^-1 U_star N^-T
            for iatom in range(natom):
                new_mat[iatom] = np.matmul(ninv, np.matmul(new_mat[iatom], ninv.T))

            return new_mat

        raise ValueError("Invalid format: `%s`" % str(fmt))

    def get_dataframe(self, temp=300, fmt="cartesian", view="inequivalent", what="displ", decimals=4,
                      select_symbols=None, verbose=0) -> pd.DataFrame:
        """
        Return |pandas-DataFrame| with cartesian tensor components as columns and (inequivalent) sites along the rows.

        Args:
            temp: Temperature in Kelvin.
            fmt: "cartesian" for elements in Cartesian coordinates, "cif" for results in reduced coordinates
            view: "inequivalent" to show only inequivalent atoms. "all" for all sites.
            what: "displ" for displament, "vel" for velocity.
            decimals: Number of decimal places to round to.
                If decimals is negative, it specifies the number of positions to the left of the decimal point.
            select_symbols: String or list of strings with chemical symbols. Used to select only atoms of this type.
            verbose: Verbosity level.

        Return: |pandas-DataFrame|
        """
        # Select atoms.
        aview = self._get_atomview(view, select_symbols=select_symbols, verbose=verbose)

        # [natom, 3, 3, nt=1]
        msq = self.get_msq_tmesh([float(temp)], iatom_list=aview.iatom_list, what_list=what)
        ucart = getattr(msq, what)
        natom = len(self.structure)
        ucart = np.reshape(ucart, (natom, 3, 3))
        values = ucart
        if what == "displ":
            values = self.convert_ucart(ucart, fmt)

        columns = ["xx", "yy", "zz", "yz", "xz", "xy"]
        inds = [(0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1)]
        rows = []
        for (iatom, wlabel) in zip(aview.iatom_list, aview.wyck_labels):
            site = self.structure[iatom]
            d = OrderedDict()
            d["element"] = site.specie.symbol
            d["site_index"] = iatom
            d["frac_coords"] = np.round(site.frac_coords, decimals=decimals)
            d["cart_coords"] = np.round(site.coords, decimals=decimals)
            d["wyckoff"] = wlabel
            if fmt == "cartesian":
                d["isotropic"] = ucart[iatom].trace() / 3.0
                d["determinant"] = np.linalg.det(ucart[iatom])
            for col, ind in zip(columns, inds):
                d[col] = values[iatom, ind[0], ind[1]]
            rows.append(d)

        return pd.DataFrame(rows, columns=list(rows[0].keys()) if rows else None)

    def write_cif_file(self, filepath, temp=300, symprec=None) -> str:
        """
        Write CIF file with structure and anisotropic U tensor in CIF format.

        Args:
            filepath: Name of CIF file. If None, a temporary filepath is created.
            temp: Temperature in Kelvin (used to compute U).
            symprec (float): If not none, finds the symmetry of the structure
                and writes the cif with symmetry information. Passes symprec to the SpacegroupAnalyzer

        Return: Filepath
        """
        if filepath is None:
            import tempfile
            _, filepath = tempfile.mkstemp(suffix=".cif", text=True)

        with open(filepath, "wt") as fh:
            fh.write(self.get_cif_string(temp=temp, symprec=symprec))

        return filepath

    #def jsmol(self, temp=300, symprec=None, verbose=0): # pragma: no cover
    #    """
    #    Args:
    #        symprec (float): If not none, finds the symmetry of the structure
    #            and writes the cif with symmetry information. Passes symprec to the SpacegroupAnalyzer
    #        verbose: Verbosity level.
    #    """
    #    cif_string = self.get_cif_string(temp=temp, symprec=symprec)

    #    try:
    #      from jupyter_jsmol import JsmolView
    #    except ImportError:
    #        raise ImportError("jupyter_jsmol is not installed. See https://github.com/fekad/jupyter-jsmol")

    #    jsmol = JsmolView(color='white')
    #    from IPython.display import display, HTML
    #    FIXME TEMPORARY HACK
    #    display(HTML('<script type="text/javascript" src="/nbextensions/jupyter-jsmol/jsmol/JSmol.min.js"></script>'))
    #    display(jsmol)

    #    cmd = 'load inline "%s" {1 1 1}; ellipsoid;' % cif_string
    #    if verbose: print("executing cmd:", cmd)
    #    jsmol.script(cmd)

    #    return jsmol

    def vesta_open(self, temp=300): # pragma: no cover
        """
        Visualize termal displacement ellipsoids at temperature `temp` (Kelvin) with Vesta_ application.
        """
        filepath = self.write_cif_file(filepath=None, temp=temp)
        cprint("Writing structure + Debye-Waller tensor in CIF format for T = %s (K) to file: %s" % (temp, filepath), "green")
        cprint("In the Vesta GUI, select: Properties -> Atoms -> Show as displament ellipsoids.", "green")
        from abipy.iotools import Visualizer
        visu = Visualizer.from_name("vesta")

        return visu(filepath)()

    def get_cif_string(self, temp=300, symprec=None) -> str:
        """
        Return string with structure and anisotropic U tensor in CIF format at temperature `temp` in Kelvin

        Args:
            symprec (float): If not none, finds the symmetry of the structure
                and writes the cif with symmetry information. Passes symprec
                to the SpacegroupAnalyzer
        """
        # Get string with structure in CIF format.
        # Don't use symprec because it changes the order of the sites
        # and we must be consistent with site_labels when writing aniso_U terms!
        from pymatgen.io.cif import CifWriter
        cif = CifWriter(self.structure, symprec=symprec)

        aniso_u = """loop_
_atom_site_aniso_label
_atom_site_aniso_U_11
_atom_site_aniso_U_22
_atom_site_aniso_U_33
_atom_site_aniso_U_23
_atom_site_aniso_U_13
_atom_site_aniso_U_12""".splitlines()

        # Compute U tensor in CIF format (reduced coords)
        natom = len(self.structure)
        msq = self.get_msq_tmesh([float(temp)], what_list="displ")
        ucart = getattr(msq, "displ")
        ucart = np.reshape(ucart, (natom, 3, 3))
        ucif = self.convert_ucart(ucart, fmt="cif")

        # Add matrix elements. Use 0 based index
        for iatom, site in enumerate(self.structure):
            site_label = "%s%d" % (site.specie.symbol, iatom)
            m = ucif[iatom]
            aniso_u.append("%s %10.5f %10.5f %10.5f %10.5f %10.5f %10.5f" %
                    (site_label, m[0, 0], m[1, 1], m[2, 2], m[1, 2], m[0, 2], m[0, 1]))

        return str(cif) + "\n".join(aniso_u)

    def check_site_symmetries(self, temp=300, verbose=0):
        """
        Check site symmetries of the displacement tensor at temperature `temp` in Kelvin.
        Return maximum error.
        """
        msq = self.get_msq_tmesh([float(temp)], what_list="displ")
        ucart = getattr(msq, "displ")
        natom = len(self.structure)
        ucart = np.reshape(ucart, (natom, 3, 3))

        return self.structure.site_symmetries.check_site_symmetries(ucart, verbose=verbose)

    def _get_components(self, components):
        """
        Return list of components to analyze from user input.
        """
        if components == "all":
            return list(self.ALL_COMPS.values())
        elif components == "upper":
            return [self.ALL_COMPS[c] for c in ("xx", "yy", "zz", "yz", "xz", "xy")]
        elif components == "diag":
            return [self.ALL_COMPS[c] for c in ("xx", "yy", "zz")]
        elif components == "offdiag":
            return [self.ALL_COMPS[c] for c in ("xy", "xz", "yz")]
        else:
            return [self.ALL_COMPS[c] for c in list_strings(components)]

    @add_fig_kwargs
    def plot(self, components="upper", view="inequivalent", units="eV", select_symbols=None,
             xlims=None, ylims=None, sharey=False, fontsize=8, verbose=0, **kwargs):
        """
        Plot the generalized phonon DOS g_ij(omega, atom) for each atom in the unit cell.
        One subplot per atom. Each subplot shows the 9 independent components of the symmetric tensor.
        as a function of frequency. By default, only "inequivalent" atoms are shown.

        Args:
            view: "inequivalent" to show only inequivalent atoms. "all" for all sites.
            components: List of cartesian tensor components to plot e.g. ["xx", "xy"].
                "all" for all components. "upper" for the upper triangle, "diag" for diagonal elements.
            units: Units energy axis. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz").
                Case-insensitive.
            select_symbols: String or list of strings with chemical symbols. Used to select only atoms of this type.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.
            ylims: Set the data limits for the y-axis.
            sharey: True if y-axis should be shared.
            fontsize: Legend and title fontsize.
            verbose: Verbosity level.

        Returns: |matplotlib-Figure|
        """
        # TODO Decide units for internal arrays.
        factor = abu.phfactor_ev2units(units)

        # Select atoms.
        aview = self._get_atomview(view, select_symbols, verbose=verbose)

        num_plots = len(aview.iatom_list)
        nrows, ncols = 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = num_plots // ncols + num_plots % ncols

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=sharey, squeeze=True)
        ax_list = np.reshape(ax_list, (nrows, ncols)).ravel()
        # don't show the last ax if num_plots is odd.
        if num_plots % ncols != 0: ax_list[-1].axis("off")

        xx = self.wmesh * factor
        components = self._get_components(components)

        # For each atom in the view.
        for ix, (ax, iatom, site_label) in enumerate(zip(ax_list, aview.iatom_list, aview.site_labels)):
            irow, icol = divmod(ix, ncols)
            ax.grid(True)
            set_axlims(ax, xlims, "x")
            set_axlims(ax, ylims, "y")
            ax.set_title(site_label, fontsize=fontsize)
            #site = self.structure[iatom]
            #color = cmap(float(iatom) / max((len(iatom_list) - 1), 1))

            # Plot components for this atom on the same ax.
            for c in components:
                yw = c.eval33w(self.values[iatom])
                label = r"$G_{%s}$" % c.name if ix == 0 else None
                ax.plot(xx, yw / factor, label=label, **c.plot_kwargs)

            # Handle labels.
            if irow == nrows - 1:
                ax.set_xlabel('Frequency %s' % abu.phunit_tag(units))
            else:
                set_visible(ax, False, "xlabel", "xticklabels")

            if ix == 0:
                ax.set_ylabel(r"$g_{ij}(\omega)$ 1/%s (Cart coords)" % abu.phunit_tag(units))
                ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    @add_fig_kwargs
    def plot_tensor(self, tstart=0, tstop=600, num=50, components="all", what="displ", view="inequivalent",
                    select_symbols=None, colormap="jet", xlims=None, ylims=None, fontsize=10, verbose=0, **kwargs):
        """
        Plot tensor(T) for each atom in the unit cell.
        One subplot for each component, each subplot show all inequivalent sites.
        By default, only "inequivalent" atoms are shown.

        Args:
            tstart: The starting value (in Kelvin) of the temperature mesh.
            tstop: The end value (in Kelvin) of the mesh.
            num: int, optional Number of samples to generate.
            components: "all" for all components. "diag" for diagonal elements, "offdiag" for off-diagonal terms only.
            what: "displ" for displament, "vel" for velocity.
            view: "inequivalent" to show only inequivalent atoms. "all" for all sites.
            select_symbols: String or list of strings with chemical symbols. Used to select only atoms of this type.
            colormap: matplotlib colormap.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used
            fontsize: Legend and title fontsize.
            verbose: Verbosity level.

        Returns: |matplotlib-Figure|
        """
        # Select atoms.
        aview = self._get_atomview(view, select_symbols=select_symbols, verbose=verbose)

        # One subplot for each component
        diag = ["xx", "yy", "zz"]
        offdiag = ["xy", "xz", "yz"]
        components = {
            "all": diag + offdiag, "diag": diag, "offdiag": offdiag,
        }[components]

        components = self._get_components(components)
        shape = np.reshape(components, (-1, 3)).shape
        nrows, ncols = shape[0], shape[1]

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=True, squeeze=True)
        ax_list = np.reshape(ax_list, (nrows, ncols)).ravel()
        cmap = plt.get_cmap(colormap)

        # Compute U(T)
        tmesh = np.linspace(tstart, tstop, num=num)
        msq = self.get_msq_tmesh(tmesh, iatom_list=aview.iatom_list, what_list=what)
        # [natom,3,3,nt] array
        values = getattr(msq, what)

        for ix, (ax, comp) in enumerate(zip(ax_list, components)):
            irow, icol = divmod(ix, ncols)
            ax.grid(True)
            set_axlims(ax, xlims, "x")
            set_axlims(ax, ylims, "y")
            ylabel = comp.get_tavg_label(what, with_units=True)
            ax.set_ylabel(ylabel, fontsize=fontsize)

            # Plot this component for all inequivalent atoms on the same subplot.
            for ii, (iatom, site_label) in enumerate(zip(aview.iatom_list, aview.site_labels)):
                color = cmap(float(ii) / max((len(aview.iatom_list) - 1), 1))
                ys = comp.eval33w(values[iatom])
                ax.plot(msq.tmesh, ys, label=site_label if ix == 0 else None,
                        color=color) #, marker="o")
                if ix == 0:
                    ax.legend(loc="best", fontsize=fontsize, shadow=True)

            if irow == 1:
                ax.set_xlabel('Temperature (K)')
            else:
                set_visible(ax, False, "xlabel", "xticklabels")

        return fig

    @add_fig_kwargs
    def plot_uiso(self, tstart=0, tstop=600, num=50, what="displ", view="inequivalent",
                  select_symbols=None, colormap="jet", xlims=None, ylims=None, sharey=False,
                  fontsize=10, verbose=0, **kwargs):
        """
        Plot phonon PJDOS for each atom in the unit cell.
        One subplot for each component, each subplot show all inequivalent sites.
        By default, only "inequivalent" atoms are shown.

        comparison of Ueq values, which
        are calculated as the mean of the diagonal elements of the harmonic ADP tensor, (d)
        comparison of the ADP anisotropy factor, which is defined as the ratio between maximum Uii
        and minimum Uii values. A ratio of 1 would correspond to an isotropic displacement.

        Args:
            tstart: The starting value (in Kelvin) of the temperature mesh.
            tstop: The end value (in Kelvin) of the mesh.
            num: int, optional Number of samples to generate.
            components: "all" for all components. "diag" for diagonal elements, "offdiag" for off-diagonal terms only.
            what: "displ" for displament, "vel" for velocity.
            view: "inequivalent" to show only inequivalent atoms. "all" for all sites.
            select_symbols: String or list of strings with chemical symbols. Used to select only atoms of this type.
            colormap: matplotlib colormap.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used
            sharey: True if y-axis should be shared.
            fontsize: Legend and title fontsize.
            verbose: Verbosity level.

        Returns: |matplotlib-Figure|
        """
        # Select atoms.
        aview = self._get_atomview(view, select_symbols=select_symbols, verbose=verbose)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=2, ncols=1,
                                                sharex=True, sharey=sharey, squeeze=True)
        cmap = plt.get_cmap(colormap)

        # Compute U(T)
        tmesh = np.linspace(tstart, tstop, num=num)
        msq = self.get_msq_tmesh(tmesh, iatom_list=aview.iatom_list, what_list=what)
        # [natom, 3, 3, nt]
        values = getattr(msq, what)
        ntemp = len(msq.tmesh)

        for ix, ax in enumerate(ax_list):
            ax.grid(True)
            set_axlims(ax, xlims, "x")
            set_axlims(ax, ylims, "y")
            if what == "displ":
                ylabel = r"$U_{iso}\;(\AA^2)$" if ix == 0 else \
                         r"Anisotropy factor ($\dfrac{\epsilon_{max}}{\epsilon_{min}}}$)"
            elif what == "vel":
                ylabel = r"$V_{iso}\;(m/s)^2$" if ix == 0 else \
                         r"Anisotropy factor ($\dfrac{\epsilon_{max}}{\epsilon_{min}}}$)"
            else:
                raise ValueError("Unknown value for what: `%s`" % str(what))
            ax.set_ylabel(ylabel, fontsize=fontsize)

            # Plot this component for all inequivalent atoms on the same subplot.
            for ii, (iatom, site_label) in enumerate(zip(aview.iatom_list, aview.site_labels)):
                color = cmap(float(ii) / max((len(aview.iatom_list) - 1), 1))
                #msq.displ[iatom, 3, 3, nt]
                if ix == 0:
                    # ISO calculated as the mean of the diagonal elements of the harmonic ADP tensor
                    ys = np.trace(values[iatom]) / 3.0
                elif ix == 1:
                    # Ratio between maximum Uii and minimum Uii values.
                    # A ratio of 1 would correspond to an isotropic displacement.
                    ys = np.empty(ntemp)
                    for itemp in range(ntemp):
                        eigs = np.linalg.eigvalsh(values[iatom, :, :, itemp], UPLO='U')
                        ys[itemp] = eigs.max() / eigs.min()
                else:
                    raise ValueError("Invalid ix index: `%s" % ix)

                ax.plot(msq.tmesh, ys, label=site_label if ix == 0 else None,
                        color=color)  #, marker="o")
                if ix == 0:
                    ax.legend(loc="best", fontsize=fontsize, shadow=True)

            if ix == len(ax_list) - 1:
                ax.set_xlabel("Temperature (K)")
            else:
                set_visible(ax, False, "xlabel", "xticklabels")

        return fig



#!/usr/bin/env python
r"""
Relaxation with different k-meshes
==================================

In this example, we employ the relaxation algorithms implemented in Abinit (``ionmov`` and ``optcell``)
to find the equilibrium configuration of GaN (atomic positions and lattice vectors).
The relaxation is performed with different k-meshes to monitor the convergence of the results.
You will observe a change of the equilibrium parameters with respect to the k-point mesh.

Note the we are using pseudopotentials generated with the GGA which tends to
overestimate the lattice parameters and ecut is way too low.
If you replace GGA with LDA, you will observe that LDA tends to underestimate the parameters.
"""

import sys
import os

import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata


def build_flow(options):
    """
    Build and return a flow performing structural relaxations with different k-point samplings.
    """
    # Set working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # List of k-meshes.
    ngkpt_list = [
        [3, 3, 2],
        [6, 6, 4],
        [8, 8, 6],
    ]

    structure = abilab.Structure.from_file(abidata.cif_file("gan2.cif"))
    pseudos = abidata.pseudos("Ga.oncvpsp", "N.oncvpsp")

    # Build multidataset.
    multi = abilab.MultiDataset(structure=structure, pseudos=pseudos, ndtset=len(ngkpt_list))

    # Set global variables for structural relaxation. Note dilatmx and ecutsm
    # Ecut should depend on pseudos.
    multi.set_vars(
        ecut=15,       # Too low
        optcell=2,
        ionmov=3,
        tolrff=5.0e-2,
        tolmxf=5.0e-5,
        ntime=50,
        dilatmx=1.05,  # Important!
        ecutsm=0.5,    # Important!
    )

    # Here we set the k-meshes (Gamma-centered for simplicity)
    for i, ngkpt in enumerate(ngkpt_list):
        multi[i].set_kmesh(ngkpt=ngkpt, shiftk=[0, 0, 0])

    # As the calculations are independent, we can use Flow.from_inputs
    # and call split_datasets to create len(ngkpt_list) inputs.
    # Note that it's a good idea to specify the task_class so that AbiPy knows how to restart the calculation.
    return flowtk.Flow.from_inputs(options.workdir, inputs=multi.split_datasets(),
                                   task_class=flowtk.RelaxTask)


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())

############################################################################
# Run the script with:
#
#     run_relax_vs_kpts.py -s
#
# then use:
#
#     abirun.py flow_relax_vs_kpts hist -p
#
# to print (and plot) the relaxed parameters at the end of the run.
#
# .. code-block:: bash
#
#     Table with final structures, pressures in GPa and force stats in eV/Ang:
#
#           formula  natom  angle0  angle1  angle2      a      b      c  volume  \
#     w0_t0  Ga2 N2      4    90.0    90.0   120.0  3.224  3.224  5.343  48.080
#     w0_t1  Ga2 N2      4    90.0    90.0   120.0  3.249  3.249  5.321  48.635
#     w0_t2  Ga2 N2      4    90.0    90.0   120.0  3.254  3.254  5.335  48.909
#
#           abispg_num  num_steps  final_energy  final_pressure task_class  \
#     w0_t0       None          9      -755.344      -1.088e-04  RelaxTask
#     w0_t1       None          9      -755.632      -5.088e-03  RelaxTask
#     w0_t2       None         11      -755.645      -4.314e-03  RelaxTask
#
#                                                  ncfile              status
#     w0_t0  flow_relax_vs_kpts/w0/t0/outdata/out_HIST.nc  Completed
#     w0_t1  flow_relax_vs_kpts/w0/t1/outdata/out_HIST.nc  Completed
#     w0_t2  flow_relax_vs_kpts/w0/t2/outdata/out_HIST.nc  Completed
#
# The experimental results are:
#
#     * Volume of the unit cell of GaN: 45.73 A^3
#     * Lattice parameters of GaN: a = 3.190 A, c = 5.189 A
#     * Vertical distance between Ga and N : about 0.377 * c [ Schulz & Thiemann, 1977]
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_relax_vs_kpts.png?raw=true
#    :alt: Evolution of the volume during the relaxation algorithm.



#!/usr/bin/env python
from __future__ import annotations

#import sys
import os
import functools
import itertools
import param
import panel as pn
import pandas as pd
import panel.widgets as pnw

from monty.termcolor import cprint
from monty.string import list_strings
from abipy.panels.core import AbipyParameterized, depends_on_btn_click, mpl, dfc, ply, ButtonContext, Loading
from abipy.tools.numtools import build_mesh
from abipy.ppcodes.ppgen import OncvGenerator
#from abipy.ppcodes.oncv_parser import OncvParser


GE_ANNOTATED = """
# Copied from ~oncvpsp/doc/32_Ge_annotated.dat

# ATOM AND REFERENCE CONFIGURATION
#
### atsym  atomic symbol
### z  atomic number
### nc number of core states
### nv number of valence states
### iexc  exchange-correlation functional: 1-Wigner, 2-Hedin-Lundquist,
###  3-Perdew-Zunger-Ceperly-Alder, 4-Perdew-Burke-Enzerhof
### psfile format of pseudopotential file, psp8 for ABINIT, upf for PWSCF
#
# atsym, z, nc, nv, iexc   psfile
    Ge  32.0   5   3   3   psp8
#
#
### n principal quantum number
### l angular momentum
### f occupancy (MUST be >0)
#
# n, l, f  (nc+nv lines)
    1    0    2.0
    2    0    2.0
    2    1    6.0
    3    0    2.0
    3    1    6.0
    3    2   10.0
    4    0    2.0
    4    1    2.0
#
# PSEUDOPOTENTIAL AND OPTIMIZATION
#
###lmax maximum angular momentum for which psp is calculated (<=3)
#
# lmax
    2
#
#
### l angular momentum
### rc  core radius for this l
### ep  energy at which psp is generated (eigenvalue inserted for occupied
###     state in reference configuration, positive energy must be specified
###     for barrier-confined "scattering" state for unoccupied l <=lmax
###     A small positive energy is usually  good (0.1-0.25 Ha).
### ncon number of constraints for pseudo wave function to match all-
###      electron wave function at rc, value + ncon-1 derivatives,
###      must be between 3 and 5 ("M" in the paper, Eq.(6))
### nbas number of basis functions.  Must be between ncon+2 and ncon+5
###      ("N" in the paper, Eqs.(4-5))
### qcut wave vector defining "residual energy" in the RRKJ method
###      ("q_c" in the paper, Eq.(1)
#
# l, rc, ep, ncon, nbas, qcut  (lmax+1 lines, l's must be in order)
    0    2.60   -0.00    4    8    5.00
    1    2.60   -0.00    4    8    5.20
    2    2.00    0.00    4    9    8.40
#
# LOCAL POTENTIAL
#
### lloc angular momentum whose semi-local psp is taken as local.  lloc=4
###      denotes a smooth polynomial continuation of the all-electron
###      potential.  If lloc<=lmax, remaining data are ignored, but
###      must be there (zeros are OK).  The rc corresponding to lloc
###      MUST BE THE MINIMUM rc for lloc<=lmax, or <= the minumum for
###      lloc=4 (usually the best choice)
### lpopt type of polynomial continuation for lloc=4. values 1-5
###       permitted.
###     1) match 2 derivatives, r^0,2,4
###     2) match 2 derivatives, r^0,4,6
###     3) match 3 derivatives, r^0,4,5,6
###     4) match 3 derivatives, r^0,4,6,8
###     5) match 3 derivatives, r^0,2,4,6
### dvloc0 shift of r=0 potential from basic continuation (Ha) for lloc=4
###        depends on lpopt above as follows, with x=(r/rc)
###        1) dvloc0*(1-x^2)^3
###        2) dvloc0*(1-x^4)^3
###        3-5) dvloc0*(1-x^4)^4
#
# lloc, lpopt, rc(5), dvloc0
    4    5    2.0    0.0
#
# VANDERBILT-KLEINMAN-BYLANDER PROJECTORs
#
### l angular momentum
### nproj number of projectors, 1 or 2.  automatically set to 0 for l=lloc
### debl energy  added to basic psp  energy ep for 2nd projector
###      automatically reset to match 2nd projector with 2nd bound state
###      at this l when it is occupied (ie., the psp is generated for a
###      correcponding-l shallow core state)
#
# l, nproj, debl  (lmax+1 lines, l's in order)
    0    2    1.50
    1    2    1.50
    2    2    1.50
#
# MODEL CORE CHARGE
#
### icmod 0: no non-linear core correction charge. 1: smooth monotonic
###       polynomial model core charge fit at "matching" rc following
###       reference 35. For icmod = 2, 3, 4 see doc/core_correction.txt
### fcfact radius for above determined by  rho_core(r)=fcfact*rho_pseudo_
###        valence(r) values 0.25-0.5 are usually good (look at plots)
###       For icmod = 3, fcfact has a different meaning and a third
###       artument rcfact is added (see core_correction.txt)
###       For icmod = 4,, fcfact is ignored but must be present (0.0 OK)
#
# icmod, fcfact, (rcfact)
    0    0.25
#
# LOG DERIVATIVE ANALYSIS
#
### epsh1 lower energy limit for "phase-shift-like" log-derivative plots
###       should be below the nearest core level for this l to make sure
###       there are no ghosts, but -2.0 Ha usually is OK  When semi-cores
###       are treated as valence, this should be below the lowest core
###       energy
### epsh2 upper energy limit, 2.0 usually good
### depsh energy mesh interval for plot, 0.02 usually good enough
#
# epsh1, epsh2, depsh
   -2.0  2.0  0.02
#
# OUTPUT GRID
#
### rlmax maximum radius for Abinit psp code 8 format output.  must be
###       greater than maximum rc (including lloc=4 rc), but also determines
###       range of diagnostic plots, so  ~2-3*rcmax is usually good
### drl mesh spacing of linear radial mesh for Abinit output. 0.02 is good
###     for "softer" psps, 0.01 is probably better with 1st row, 3d's, or
###     semi-core psps.
#
# rlmax, drl
    4.0  0.01
#
# TEST CONFIGURATIONS
#
### ncnf number of test configurations (<=4)  The reference config is always
###      run first as a consistency check.  core always is the reference core
###      configuration.  The excitation energy of the configuration for the
###      pseudo-atom will be compared with the all-electron result.
# ncnf
    4
#
### n principal quantum number for all-electron atom
### l angular momentum
### f occupancy
#
#   nvcnf (repeated ncnf times)
#         number of valence states in this test configuration
#   n, l, f  (nvcnf lines, repeated follwing nvcnf's ncnf times)
#   n    l    f
    3
    3    2   10.00
    4    0    1.00
    4    1    2.00
#
    3
    3    2   10.00
    4    0    2.00
    4    1    1.00
#
    3
    3    2   10.00
    4    0    1.00
    4    1    1.00
#
    3
    3    2   10.00
    4    0    1.00
    4    2    1.00
#
"""

class Lparams(AbipyParameterized):
    """
    Stores all the oncvpsp pseudization parameters for a given l.
    """

    l = param.Integer(None, bounds=(0, None))
    rc = param.Number(None, bounds=(0, None))
    ep = param.Number(None, bounds=(None, None))
    ncon = param.Integer(None, bounds=(0, 9))
    nbas = param.Integer(None, bounds=(0, 9))
    qcut = param.Number(None, bounds=(0, None))
    nproj = param.Integer(None, bounds=(1, 5,))
    debl = param.Number(None, bounds=(None, None))



class Nlf(AbipyParameterized):
    """
    Stores the value of n, l and occupancy f.
    """
    n = param.Integer(None, bounds=(1, None))
    l = param.Integer(None, bounds=(0, None))
    f = param.Number(None, bounds=(0.0, None))



class OncvInput(AbipyParameterized):
    """
    A parametrized class with all the Oncvpsp input variables,
    usually constructed from an external file.
    """

    atsym = param.String()
    z = param.Integer(1, bounds=(1, None))
    nc = param.Integer(0, bounds=(0, None))
    nv = param.Integer(0, bounds=(0, None))
    iexc = param.Integer(0)
    psfile = param.String("both")

    nlf_list = param.List() #[], item_type=Nlf)
    lmax = param.Integer(0, bounds=(0, 4))
    lparams = param.List() #[], item_type=Lparams)

    lloc = param.Integer(4)
    lpopt = param.Integer(5)
    rc5 = param.Number(0.0)
    dvloc0 = param.Number(0.0)

    icmod = param.Integer(0, bounds=(0, 3))
    fcfact = param.Number(0.0, bounds=(0, None))
    rcfact = param.Number(0.0, bounds=(0, None))

    epsh1 = param.Number(-25)
    epsh2 = param.Number(+25)
    depsh = param.Number(0.02)
    rlmax = param.Number(6.0)
    drl = param.Number(0.1)

    nconf = param.Integer(0, bounds=(0, None))

    @classmethod
    def from_file(cls, path: str) -> OncvInput:
        """
        Initialize the object from file.
        """
        with open(path, "rt") as fh:
            return cls.from_string(fh.read())

    @classmethod
    def from_string(cls, string: str) -> OncvInput:
        """Build the object from a string."""

        def parse_line(line: str, types: str):
            """
            Helper function to parse line, convert each token according to
            character in `types` and return list of values or value dependening on num char in types.
            s -> str, i -> int, f -> float.
            """
            try:
                i = line.find("#")
                if i != -1:
                    line = line[:i]
                tokens = line.split()

                if len(tokens) != len(types):
                    raise ValueError(f"Expecting {len(types)} tokens but got {len(tokens)} tokens: `{tokens}`")

                outs = []
                for tok, typ in zip(tokens, types):
                    if typ == "s":
                        outs.append(str(tok))
                    elif typ == "i":
                        fval = float(tok)
                        ival = int(fval)
                        if fval != ival:
                            raise TypeError(f"Expecting int in line {line}, got : `{tok}`")
                        outs.append(ival)
                    elif typ == "f":
                        outs.append(float(tok))
                    else:
                        raise TypeError(f"Invalid type: `{typ}`")

                return outs if len(outs) > 1 else outs[0]

            except Exception as exc:
                print(exc)
                raise ValueError(f"Invalid line: `{line}`, expecting types: `{types}`")

        lines = [l for l in string.split("\n") if not l.startswith("#")]
        lines = [l for l in lines if l.strip()]

        # atsym z nc nv iexc psfile
        atsym, z, nc, nv, iexc, psfile = parse_line(lines.pop(0), "siiiis")
        psfile = "both"

        nlf_list = []
        for i in range(nc + nv):
            # n l f
            n, l, f = parse_line(lines.pop(0), "iif")
            nlf_list.append(Nlf(n=n, l=l, f=f))

        lmax = parse_line(lines.pop(0), "i")

        params_l = {}
        for i in range(lmax + 1):
            # l rc ep ncon nbas qcut
            l, rc, ep, ncon, nbas, qcut = parse_line(lines.pop(0), "iffiif")
            assert i == l
            params_l[l] = dict(l=l, rc=rc, ep=ep, ncon=ncon, nbas=nbas, qcut=qcut)

        # lloc lpopt rc5 dvloc0
        lloc, lpopt, rc5, dvloc0 = parse_line(lines.pop(0), "iiff")

        for l in range(lmax + 1):
            # l nproj debl
            l, nproj, debl = parse_line(lines.pop(0), "iif")
            params_l[l].update(nproj=nproj, debl=debl)

        lparams = [Lparams(**d) for d in params_l.values()]

        # icmod fcfact [rcfact]
        l = lines.pop(0)
        rcfact = 0.0
        try:
            icmod, fcfact, rcfact = parse_line(l, "iff")
        except ValueError:
            icmod, fcfact = parse_line(l, "if")

        # epsh1 epsh2 depsh
        epsh1, epsh2, depsh = parse_line(lines.pop(0), "fff")
        # rlmax drl
        rlmax, drl = parse_line(lines.pop(0), "ff")
        #nconf = parse_line(lines.pop(0), "i")
        nconf = 0

        locs = locals()
        d = {k: locs[k]for k in [
            "atsym",
            "z",
            "nc",
            "nv",
            "iexc",
            "psfile",
            "nlf_list",
            "lparams",
            "lmax",
            "lloc",
            "lpopt",
            "rc5",
            "dvloc0",
            "icmod",
            "fcfact",
            "rcfact",
            "epsh1",
            "epsh2",
            "depsh",
            "rlmax",
            "drl",
            "nconf",
        ]}

        return cls(**d)


    def __init__(self, **params):
        super().__init__(**params)

    def __str__(self) -> str:
        """String with the ONCVPSP input file."""
        lines = []
        app = lines.append

        app("# ATOM AND REFERENCE CONFIGURATION")
        app("# atsym z nc nv iexc psfile")
        app(f"{self.atsym} {self.z} {self.nc} {self.nv} {self.iexc} {self.psfile}")
        app("# n l f (nc+nv lines)")
        for i, nlf in enumerate(self.nlf_list):
            tag = "# end core" if i + 1 == self.nc else ""
            app(f"{nlf.n} {nlf.l} {nlf.f} {tag}")

        app("# PSEUDOPOTENTIAL AND OPTIMIZATION")
        app("# lmax")
        app(f"{self.lmax}")
        app("# l rc ep ncon nbas qcut (lmax+1 lines, l's must be in order)")
        for p in self.lparams:
            app(f"{p.l} {p.rc:.2f} {p.ep:.2f} {p.ncon} {p.nbas} {p.qcut:.2f}")
        app("# LOCAL POTENTIAL")
        app("# lloc lpopt rc5 dvloc0")
        app(f"{self.lloc} {self.lpopt} {self.rc5:.2f} {self.dvloc0}")
        app("# VANDERBILT-KLEINMAN-BYLANDER PROJECTORs")
        app("# l nproj debl (lmax+1 lines, l's in order")
        for p in self.lparams:
            app(f"{p.l} {p.nproj} {p.debl:.2f}")

        app("# MODEL CORE CHARGE")
        app("# icmod fcfact rcfact")
        app(f"{self.icmod} {self.fcfact:.2f} {self.rcfact:.2f}")
        app("# LOG DERIVATIVE ANALYSIS")
        app("# epsh1 epsh2 depsh")
        app(f"{self.epsh1} {self.epsh2} {self.depsh}")
        app("# OUTPUT GRID")
        app("# rlmax drl")
        app(f"{self.rlmax} {self.drl}")
        app("# TEST CONFIGURATIONS")
        app("# ncnf")
        app(f"{self.nconf}")

        return "\n".join(lines)

    def get_min_rc(self) -> float:
        """Return the minimum of rc(l) over l."""
        return min(p.rc for p in self.lparams)

    def find_lparam(self, l: int, what: str) -> tuple(int, float):
        for i, p in enumerate(self.lparams):
            if p.l == l:
                return i, getattr(p, what)

        raise ValueError(f"Cannot find l: {l} in input with what: {what}")


def run_psgen(psgen: OncvGenerator, data: dict) -> dict:
    """
    Helper function to run oncvpsp, parse the outputfiles and return dictionary with results

    Args:
        psgen (OncvGenerator): [description]
        data (dict): input dictionary with extra data to be added to the output dict
    """
    # Init return values
    max_ecut = None
    max_atan_logder_l1err = None
    max_psexc_abserr = None
    herm_err = None
    status = None
    nwarns = None
    nerrs = None

    try:
        psgen.start()
        retcode = psgen.wait()

        results = psgen.parser.get_results()

        if results is not None:
            max_ecut = results.max_ecut
            max_atan_logder_l1err = results.max_atan_logder_l1err
            max_psexc_abserr = results.max_psexc_abserr
            herm_err = results.herm_err
            status = str(psgen.status)
            nwarns = len(psgen.parser.warnings)
            nerrs = len(psgen.parser.errors)

    except Exception as exc:
        print("Exception in run_psgen", exc)

    d =  dict(
        status=status,
        max_ecut=max_ecut,
        max_atan_logder_l1err=max_atan_logder_l1err,
        max_psexc_abserr=max_psexc_abserr,
        herm_err=herm_err,
        nwarns=nwarns,
        nerrs=nerrs,
    )

    data = data.copy()
    data.update(**d)
    return data

class OncvGui(AbipyParameterized):
    calc_type = param.ObjectSelector(default="scalar-relativistic",
                                     objects=["scalar-relativistic", "fully-relativistic", "non-relativistic"],
                                     label="Relativistic effects")

    max_nprocs =  param.Integer(max(os.cpu_count() // 2, 1), bounds=(1, None))

    dpi = param.Integer(82, bounds=(24, None))

    #in_filepath = param.String("", doc="The path to the oncvps input file.")

    qcut_num =  param.Integer(2, bounds=(1, None))
    qcut_step = param.Number(1, bounds=(0, None))
    qcut_dir = param.Selector(["centered", ">", "<"])

    rc_num =  param.Integer(2, bounds=(1, None))
    rc_step = param.Number(0.1, bounds=(0, None))
    rc_dir = param.Selector(["centered", ">", "<"])

    debl_num =  param.Integer(2, bounds=(1, None))
    debl_step = param.Number(1.0, bounds=(0, None))
    debl_dir = param.Selector(["centered", ">", "<"])

    rc5_num =  param.Integer(2, bounds=(1, None))
    rc5_step = param.Number(0.1, bounds=(0, None))
    rc5_dir = param.Selector(["<", "centered", ">"])

    dvloc0_num =  param.Integer(2, bounds=(1, None))
    dvloc0_step = param.Number(0.5, bounds=(0, None))
    dvloc0_dir = param.Selector(["centered", "<", ">"])

    fcfact_num =  param.Integer(1, bounds=(1, None))
    fcfact_step = param.Number(0.2, bounds=(0, None))
    fcfact_dir = param.Selector(["centered", ">", "<"])

    rcfact_num =  param.Integer(1, bounds=(1, None))
    rcfact_step = param.Number(0.2, bounds=(0, None))
    rcfact_dir = param.Selector(["centered", ">", "<"])

    ace_theme = param.ObjectSelector(default="chrome",
                                    objects=pnw.Ace.param.theme.objects,
                                    doc="Theme of the editor")

    history_idx = param.Integer(default=-1, label="History index")

    @classmethod
    def from_file(cls, path: str, plotlyFlag: bool) -> OncvGui:
        """
        Build an instance from a file with the oncvpsp input variables.
        """
        return cls(oncv_input=OncvInput.from_file(path), plotlyFlag=plotlyFlag, in_filepath=path)

    def __init__(self, oncv_input, plotlyFlag, in_filepath="", **params):
        super().__init__(**params)

        self.ace_kwargs = dict(sizing_mode='stretch_both', print_margin=False, language='text', height=600,
                            theme="chrome",
                            #theme="dracula",
                            #max_length=150,
                            )

        self.input_ace = pnw.Ace(value=str(oncv_input), **self.ace_kwargs)

        # Add annotated example for documentation purposes.
        self.annotated_example = pn.pane.HTML(f"<pre><code> {GE_ANNOTATED} </code></pre>")

        self.in_filepath = pnw.TextInput(value=in_filepath, placeholder='Enter the filepath...')

        self.out_area = pn.Column("## Oncvpsp output:",  sizing_mode="stretch_width")
        #self.out_runtests = pn.Column("## Basic tests:", sizing_mode="stretch_width")

        # Define buttons
        self.execute_btn = pnw.Button(name="Execute", button_type='primary')
        self.execute_btn.on_click(self.on_execute_btn)

        # This is the directory used to run oncvpsp when the user clicks execute_btn
        #self._execute_stdout_path = None

        # List storing all the inputs.
        self.input_history = []
        self.history_btn = pnw.Button(name="Compare", button_type='primary')
        #self.history_btn.on_click(self.on_history_btn)

        self.rc_qcut_btn = pnw.Button(name="Execute", button_type='primary')
        
        self.plotlyFlag = plotlyFlag

    @param.depends("ace_theme")
    def change_ace_theme(self):
        #print("Changing theme")
        self.input_ace.theme = self.ace_theme

    def get_oncv_input(self) -> OncvInput:
        """
        Take the string from the ACE editor and build an oncv input
        with the last changes done by the user
        """
        return OncvInput.from_string(self.input_ace.value)

    def set_oncv_input_string(self, new_string: str) -> OncvInput:
        """
        Update the string in the ACE editor, store previous string in history.
        Return: OncvInput instance.
        """
        #print("Updating input with new_string:\n", new_string)
        self.input_history.append(self.input_ace.value)
        self.input_ace.value = new_string

        return self.get_oncv_input()

    def starmap(self, func, list_of_args):
        import time
        time_start = time.time()

        # Don't use more procs than tasks.
        _max_nprocs_ = self.max_nprocs
        _max_nprocs_ = min(_max_nprocs_, len(list_of_args))

        if _max_nprocs_ == 1:
            values = [func(*args) for args in list_of_args]
        else:
            # TODO: Optimize, use better algo
            # This Pool uses threads instead of multiprocessing
            # Cannot use multiprocessing because we have side effects in psgen
            from multiprocessing.dummy import Pool
            with Pool(processes=self.max_nprocs) as pool:
                values = pool.starmap(func, list_of_args)

        print(f"Done {len(list_of_args)} tasks in {time.time() - time_start:.2f} [s] with {_max_nprocs_} processe(s)")
        return values

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets"""
        d = {}

        main = pn.Column(
            pn.Row(
                self.pws_col(["calc_type", "max_nprocs",
                                "dpi", "ace_theme", "execute_btn"]),
                self.input_ace,
            ),
            pn.Card(self.annotated_example, title='Annotated example', collapsed=True,
                    header_color="blue", sizing_mode="stretch_width"),
            sizing_mode="stretch_width",
        )

        d["Main"] =  pn.Column(main,
                               #pn.layout.Divider(),
                               self.out_area,
                               #self.out_runtests,
                               sizing_mode="stretch_width",
                               )

        d["Rc_qcut_opt"] = self.get_rc_qcut_opt_view()
        d["History"] = self.get_history_view()

        if as_dict: return d
        template = self.get_template_from_tabs(d, template=kwargs.get("template", None))
        #self.tabs = template
        #print(self.tabs)
        return template

    def get_history_view(self) -> pn.Row:
        return pn.Row(
            self.pws_col(["## History",
                            "history_idx",
                            "history_btn",
                            ]),
            self.on_history_btn
        )

    @depends_on_btn_click('history_btn')
    def on_history_btn(self) -> pn.Column:
        #print("hello")
        hist_len = len(self.input_history)
        idx = self.history_idx
        if hist_len == 0 or idx >= hist_len:
            return pn.Column(f"hist_len == {hist_len}")

        ace_hist = pnw.Ace(value=self.input_history[idx], **self.ace_kwargs)

        fromlines = self.input_history[idx].splitlines()
        tolines = self.input_ace.value.splitlines()
        from difflib import HtmlDiff
        html_table = HtmlDiff().make_table(fromlines, tolines) #, fromdesc='', todesc='', context=False, numlines=5)

        return pn.Column(
                f"## Input at history index: {idx}",
                ace_hist,
                pn.pane.HTML(html_table),
        )

    def get_rc_widgets(self, oncv_input: OncvInput) -> pn.WidgetBox:
        """Return widgets to change the value of rc(l)"""
        menu_items = [(f"l = {l}", str(l)) for l in range(oncv_input.lmax + 1)]
        menu_button = pnw.MenuButton(name='Change rc(l)', items=menu_items, button_type='primary')
        menu_button.on_click(self.on_change_rc)
        rc_l = {p.l: p.rc for p in oncv_input.lparams}
        help_str = f"""
Here one can change the value of rc(l).

The present values of rc_l are: {rc_l}
"""
        return pn.WidgetBox(menu_button,
                            *[self.param[k] for k in ("rc_num", "rc_step", "rc_dir")],
                            help_str)

    def get_qcut_widgets(self, oncv_input: OncvInput) -> pn.WidgetBox:
        """Return widgets to change the value of qc(l)"""
        menu_items = [(f"l = {l}", str(l)) for l in range(oncv_input.lmax + 1)]
        menu_button = pnw.MenuButton(name='Change qcut(l)', items=menu_items, button_type='primary')
        menu_button.on_click(self.on_change_qcut)
        qc_l = {p.l: p.qcut for p in oncv_input.lparams}
        help_str = f"""
Here one can change the value of qcut(l).

The present values are: {qc_l}
"""
        return pn.WidgetBox(menu_button,
                            *[self.param[k] for k in ("qcut_num", "qcut_step", "qcut_dir")],
                            help_str)

    def get_debl_widgets(self, oncv_input: OncvInput) -> pn.WidgetBox:
        """Return widgets to change the value of debl(l)"""
        menu_items = [(f"l = {l}", str(l)) for l in range(oncv_input.lmax + 1)]
        menu_button = pnw.MenuButton(name='Change debl(l)', items=menu_items, button_type='primary')
        menu_button.on_click(self.on_change_debl)
        help_str = f"""
Here one can change the value of debl(l) with fixed nproj(l).
"""
        return pn.WidgetBox(menu_button,
                            *[self.param[k] for k in ("debl_num", "debl_step", "debl_dir")],
                            help_str)

    def get_rc5_widgets(self, oncv_input: OncvInput) -> pn.WidgetBox:
        """Return widgets to change the value of rc5"""
        btn = pnw.Button(name="Run", button_type='primary')
        btn.on_click(self.on_change_rc5)
        help_str = f"""
Here one can change the value of rc5 for vloc.

The present value of rc5 is {oncv_input.rc5} and min(rc) is: {oncv_input.get_min_rc()}
"""
        return pn.WidgetBox(*[self.param[k] for k in ("rc5_num", "rc5_step", "rc5_dir")],
                            btn,
                            help_str)

    def get_dvloc0_widgets(self, oncv_input: OncvInput) -> pn.WidgetBox:
        """Return widgets to change the value of dvloc0"""
        btn = pnw.Button(name="Run", button_type='primary')
        btn.on_click(self.on_change_dvloc0)
        help_str = f"""
Here one can change the value of dvloc0 for vloc.

The present value of dvloc0 is {oncv_input.dvloc0} with lpopt: {oncv_input.lpopt}
"""
        return pn.WidgetBox(*[self.param[k] for k in ("dvloc0_num", "dvloc0_step", "dvloc0_dir")],
                            btn,
                            help_str)

    def get_rhomodel_widgets(self, oncv_input: OncvInput) -> pn.WidgetBox:
        """Return widgets to change the parameters for the model core charge"""
        btn = pnw.Button(name="Run", button_type='primary')
        btn.on_click(self.on_change_rhomodel)
        help_str = f"""
Here one can change the parameters for the model core charge.

The present value of icmod is {oncv_input.icmod} with fcfact: {oncv_input.fcfact} and rcfact: {oncv_input.rcfact}
"""
        keys = ("rcfact_num", "rcfact_step", "rcfact_dir",
                "fcfact_num", "fcfact_step", "fcfact_dir")
        return pn.WidgetBox(*[self.param[k] for k in keys],
                            btn,
                            help_str)

    def gridplot_psgens(self, psgens, titles, func_names="plot_atanlogder_econv") -> pn.GridBox:
        """
        Return a GridBox with the figures obtained by calling `plotter.func_name`
        for all the PseudoGenerators in psgens.
        """
        # Generate plots with titles by calling `func_name`.
        _m = functools.partial(mpl, with_divider=False, dpi=self.dpi)
        func_names = list_strings(func_names)
        figs = []
        for psgen, title in zip(psgens, titles):
            plotter = psgen.parser.get_plotter()
            if plotter is not None:
                for func_name in func_names:
                    plot_func = getattr(plotter, func_name)
                    figs.append(_m(plot_func(show=False, title=title, fig_close=True)))

        # Insert the figures in a GridBox.
        nfigs = len(figs)
        nrows, ncols = 1, 1
        if nfigs > 1:
            ncols = 2 if len(func_names) == 1 else len(func_names)
            nrows = nfigs // ncols + nfigs % ncols
        elif nfigs == 0:
            nrows, ncols = 0, 0

        return pn.GridBox(*figs, ncols=ncols, nrows=nrows)

    def on_change_qcut(self, event) -> None:
        """
        Change the value of qcut(l), run oncvpsp and show the results.
        """
        with ButtonContext(event.obj), Loading(self.out_area):
            # Get initial qc(l) from input.
            l = int(event.new)
            oncv_input = self.get_oncv_input()
            i0, qcut0 = oncv_input.find_lparam(l, "qcut")

            # Define list of qc to be tested and build list of OncvGenerator.
            qcut_values, _ = build_mesh(qcut0, self.qcut_num, self.qcut_step, self.qcut_dir)
            psgens = []

            try:
                for qcut in qcut_values:
                    oncv_input.lparams[i0].qcut = qcut
                    psgens.append(OncvGenerator(input_str=str(oncv_input), calc_type=self.calc_type))
            finally:
                # Restore the initial value.
                oncv_input.lparams[i0].qcut = qcut0

            # Run each generator, get results, build pandas Dataframe and show it in out_area.
            tasks = [(psgen, {"qcut": qcut}) for psgen, qcut in zip(psgens, qcut_values)]
            d_list = self.starmap(run_psgen, tasks)

            dfw = self._build_table(tasks, d_list)

            head = pn.Row(pn.Column(
                            f"## Qcut optimization for l: {l}. Click the icon to update the input",
                            dfw
                            ),
                         self.get_qcut_widgets(oncv_input))

            col = pn.Column(head, sizing_mode="stretch_width")

            # Add plots
            grid = self.gridplot_psgens(psgens, [f"qc = {qc:.2f}" for qc in qcut_values])
            col.append(grid)

            self.out_area.objects = col.objects

    def _build_table(self, tasks, d_list):
        """
        Build and return a Tabulator widget with the results.
        Also, register callbacks so that it is possible to
        update the input file by clicking on the icon in the last column.
        """

        df = pd.DataFrame(d_list, columns=list(d_list[0].keys()))

        # Sort results by max_ecut and add buttons to trigger callbacks
        df = df.sort_values("max_ecut")
        dfw = pn.widgets.Tabulator(df, buttons={'accept': '<i class="fa fa-print"></i>'})

        def update_input(event):
            #print(f'Clicked {event.column!r} on row {event.row}')
            # Use the index to get the psgen as we have sorted along max_ecut
            idx = dfw.value.index[event.row]
            #print("select idx:", idx)
            psgen = tasks[idx][0]
            if event.column == "accept" and psgen.status == psgen.S_OK:
                oncv_input = self.set_oncv_input_string(psgen.input_str)
                self._update_out_area(psgen, oncv_input)

        dfw.on_click(update_input)
        return dfw

    def on_change_debl(self, event) -> None:
        """
        Change the value of debl(l), run oncvpsp and show the results.
        """
        with ButtonContext(event.obj), Loading(self.out_area):
            # Get initial qc(l) from input.
            l = int(event.new)
            oncv_input = self.get_oncv_input()

            # Define list of qc to be tested and build list of OncvGenerator.
            i0, debl0 = oncv_input.find_lparam(l, "debl")
            debl_values, _ = build_mesh(debl0, self.debl_num, self.debl_step, self.debl_dir)
            psgens = []

            try:
                for debl in debl_values:
                    oncv_input.lparams[i0].debl = debl
                    psgens.append(OncvGenerator(input_str=str(oncv_input), calc_type=self.calc_type))
            finally:
                # Restore the initial value.
                oncv_input.lparams[i0].debl = debl0

            # Run each generator, get results, build pandas Dataframe and show it in out_area.
            tasks = [(psgen, {"debl": debl}) for psgen, debl in zip(psgens, debl_values)]
            d_list = self.starmap(run_psgen, tasks)

            dfw = self._build_table(tasks, d_list)

            head = pn.Row(pn.Column(
                            f"## Debl optimization for l: {l}. Click the icon to update the input",
                            dfw,
                            ),
                         self.get_debl_widgets(oncv_input))

            col = pn.Column(head, sizing_mode="stretch_width")

            # Add plots
            grid = self.gridplot_psgens(psgens, [f"debl = {debl:.2f}" for debl in debl_values],
                                        func_names=["plot_atan_logders"])
            col.append(grid)

            self.out_area.objects = col.objects

    def on_change_rc5(self, event) -> None:
        """
        Change the value of rc5 for the local part, run oncvpsp and show the results.
        """
        with ButtonContext(event.obj), Loading(self.out_area):
            oncv_input = self.get_oncv_input()

            # Define list of qc to be tested and build list of OncvGenerator.
            rc5 = oncv_input.rc5
            rc5_values, _ = build_mesh(rc5, self.rc5_num, self.rc5_step, self.rc5_dir)
            psgens = []

            try:
                for new_rc in rc5_values:
                    oncv_input.rc5 = new_rc
                    psgens.append(OncvGenerator(input_str=str(oncv_input), calc_type=self.calc_type))
            finally:
                # Restore the initial value.
                oncv_input.rc5 = rc5

            # Run each generator, get results, build pandas Dataframe and show it in out_area.
            tasks = [(psgen, {"rc5": rc5}) for psgen, rc5 in zip(psgens, rc5_values)]
            d_list = self.starmap(run_psgen, tasks)

            dfw = self._build_table(tasks, d_list)

            head = pn.Row(pn.Column(
                            f"## Rc5 optimization. Click the icon to update the input",
                            dfw,
                            ),
                         self.get_rc5_widgets(oncv_input))

            col = pn.Column(head, sizing_mode="stretch_width")

            # Add plots
            grid = self.gridplot_psgens(psgens, [f"rc5 = {rc5:.2f}" for rc5 in rc5_values],
                                        func_names=["plot_atanlogder_econv", "plot_potentials"])
            col.append(grid)

            self.out_area.objects = col.objects

    def on_change_dvloc0(self, event) -> None:
        """
        Change the value of dvloc0 for the local part, run oncvpsp and show the results.
        """
        with ButtonContext(event.obj), Loading(self.out_area):
            oncv_input = self.get_oncv_input()

            # Define list of qc to be tested and build list of OncvGenerator.
            dvloc0 = oncv_input.dvloc0
            dvloc_values, _ = build_mesh(dvloc0, self.dvloc0_num, self.dvloc0_step, self.dvloc0_dir)
            psgens = []

            try:
                for new_dvloc in dvloc_values:
                    oncv_input.dvloc0 = new_dvloc
                    psgens.append(OncvGenerator(input_str=str(oncv_input), calc_type=self.calc_type))
            finally:
                # Restore the initial value.
                oncv_input.dvloc0 = dvloc0

            # Run each generator, get results, build pandas Dataframe and show it in out_area.
            tasks = [(psgen, {"dvloc0": dvloc}) for psgen, dvloc in zip(psgens, dvloc_values)]
            d_list = self.starmap(run_psgen, tasks)

            dfw = self._build_table(tasks, d_list)

            head = pn.Row(pn.Column(
                            f"## dvloc0 optimization. Click the icon to update the input",
                            dfw,
                            ),
                         self.get_dvloc0_widgets(oncv_input))

            col = pn.Column(head, sizing_mode="stretch_width")

            # Add plots
            grid = self.gridplot_psgens(psgens, [f"dvloc = {dvloc:.2f}" for dvloc in dvloc_values],
                                        func_names=["plot_atanlogder_econv", "plot_potentials"])
            col.append(grid)

            self.out_area.objects = col.objects

    def on_change_rc(self, event) -> None:
        """
        Change the value of rc(l), run oncvpsp and show the results.
        """
        with ButtonContext(event.obj), Loading(self.out_area):
            # Get initial rc(l) from input.
            l = int(event.new)
            oncv_input = self.get_oncv_input()

            # Define list of qc to be tested and build list of OncvGenerator.
            i0, rc0 = oncv_input.find_lparam(l, "rc")
            rc_values, _ = build_mesh(rc0, self.rc_num, self.rc_step, self.rc_dir)
            psgens = []

            try:
                for rc in rc_values:
                    oncv_input.lparams[i0].rc = rc
                    psgens.append(OncvGenerator(input_str=str(oncv_input), calc_type=self.calc_type))
            finally:
                # Restore the initial value.
                oncv_input.lparams[i0].rc = rc0

            # Run each generator, get results, build pandas Dataframe and show it in out_area.
            tasks = [(psgen, {"rc": rc}) for psgen, rc in zip(psgens, rc_values)]
            d_list = self.starmap(run_psgen, tasks)

            dfw = self._build_table(tasks, d_list)

            head = pn.Row(pn.Column(
                            f"## Rc optimization for l: {l}. Click the icon to update the input",
                            dfw,
                            ),
                         self.get_rc_widgets(oncv_input),
                         )

            col = pn.Column(head, sizing_mode="stretch_width")

            # Add plots
            grid = self.gridplot_psgens(psgens, [f"rc = {rc:.2f} for l: {l}" for rc in rc_values])
            col.append(grid)

            self.out_area.objects = col.objects

    def on_change_rhomodel(self, event) -> None:
        """
        Change the parameters for the model core charge, run oncvpsp and show the results.
        """
        with ButtonContext(event.obj), Loading(self.out_area):
            # Get initial values from input.
            oncv_input = self.get_oncv_input()
            icmod = oncv_input.icmod
            if icmod == 0: return
            fcfact0 = oncv_input.fcfact
            rcfact0 =  oncv_input.rcfact

            # Define list of values to be tested and build list of OncvGenerator.
            fcfact_values, _ = build_mesh(fcfact0, self.fcfact_num, self.fcfact_step, self.fcfact_dir)
            rcfact_values, _ = build_mesh(rcfact0, self.rcfact_num, self.rcfact_step, self.rcfact_dir)

            psgens = []
            tasks = []
            titles = []

            if icmod in (3, 4):
                # fcfact x rcfact
                for fc, rc in itertools.product(fcfact_values, rcfact_values):
                    try:
                        oncv_input.fcfact = fc
                        oncv_input.rcfact = rc
                        psgens.append(OncvGenerator(input_str=str(oncv_input), calc_type=self.calc_type))
                        tasks.append((psgens[-1], {"fcfact": fc, "rcfact": rc}))
                        titles.append(f"fcfact: {fc}, rcfact: {rc}")
                    finally:
                        # Restore the initial value.
                        oncv_input.fcfact = fcfact0
                        oncv_input.rcfact = rcfact0

            elif icmod in (1, 2):
                # Only fcfact is used here.
                for fc in fcfact_values:
                    try:
                        oncv_input.fcfact = fc
                        psgens.append(OncvGenerator(input_str=str(oncv_input), calc_type=self.calc_type))
                        tasks.append((psgens[-1], {"fcfact": fc}))
                        titles.append(f"fcfact: {fc}")
                    finally:
                        # Restore the initial value.
                        oncv_input.fcfact = fcfact0
            else:
                raise ValueError(f"Invalid icmod: {icmod}")

            # Run each generator, get results, build pandas Dataframe and show it in out_area.
            d_list = self.starmap(run_psgen, tasks)

            dfw = self._build_table(tasks, d_list)

            head = pn.Row(pn.Column(
                            f"## Rho model optimization for icmod: {icmod}. Click the icon to update the input",
                            dfw,
                            ),
                         self.get_rhomodel_widgets(oncv_input),
                         )

            col = pn.Column(head, sizing_mode="stretch_width")

            # Add plots
            grid = self.gridplot_psgens(psgens, titles,
                                       func_names=["plot_densities", "plot_den_formfact"])
            col.append(grid)

            self.out_area.objects = col.objects

    def get_rc_qcut_opt_view(self) -> pn.Row:
        oncv_input = self.get_oncv_input()

        menu_items = [(f"l = {l}", str(l)) for l in range(oncv_input.lmax + 1)]
        menu_button = pnw.MenuButton(name='Change qcut(l)', items=menu_items, button_type='primary')
        menu_button.on_click(self.on_change_rc_qcut)
        qc_l = {p.l: p.qcut for p in oncv_input.lparams}
        rc_l = {p.l: p.rc for p in oncv_input.lparams}
        help_str = f"""
Here one can change the value of rc(l) and qcut(l).

The present values of qc_l are: {qc_l}
The present values of rc_l are: {rc_l}
"""
        self.rc_qcut_out_area = pn.Column(sizing_mode="stretch_width")
        wbox = pn.WidgetBox(menu_button,
                            *[self.param[k] for k in ("qcut_num", "qcut_step", "qcut_dir")],
                            *[self.param[k] for k in ("rc_num", "rc_step", "rc_dir")],
                            help_str)

        return pn.Row(wbox, self.rc_qcut_out_area, sizing_mode="stretch_width")

    #@depends_on_btn_click('rc_qcut_btn')
    def on_change_rc_qcut(self, event) -> None:
        """
        Generate pseudos using a grid of (rc, qcut) values for given l.
        """
        with ButtonContext(event.obj), Loading(self.rc_qcut_out_area):
            # Get initial rc(l) from input.
            l = int(event.new)
            oncv_input = self.get_oncv_input()

            # Define list of qc to be tested and build list of OncvGenerator.
            i0, rc0 = oncv_input.find_lparam(l, "rc")
            rc_values, _ = build_mesh(rc0, self.rc_num, self.rc_step, self.rc_dir)

            # Define list of qc to be tested and build list of OncvGenerator.
            i0, qcut0 = oncv_input.find_lparam(l, "qcut")
            qcut_values, _ = build_mesh(qcut0, self.qcut_num, self.qcut_step, self.qcut_dir)

            def rq_prod():
                return itertools.product(rc_values, qcut_values)

            psgens = []
            try:
                for rc, qcut in rq_prod():
                    oncv_input.lparams[i0].rc = rc
                    oncv_input.lparams[i0].qcut = qcut
                    psgens.append(OncvGenerator(input_str=str(oncv_input), calc_type=self.calc_type))
            finally:
                # Restore the initial value.
                oncv_input.lparams[i0].rc = rc0
                oncv_input.lparams[i0].qcut = qcut0

            # Run each generator, get results, build pandas Dataframe and show it in out_area.
            tasks = [(psgen, {"rc": rc, "qcut": qcut}) for psgen, (rc, qcut) in zip(psgens, rq_prod())]
            d_list = self.starmap(run_psgen, tasks)

            dfw = self._build_table(tasks, d_list)

            head = pn.Row(pn.Column(
                            f"## Rc/qcut optimization for l: {l}. Click the icon to update the input",
                            dfw,
                            ),
                        )

            col = pn.Column(head, sizing_mode="stretch_width")

            # Add plots:
            grid = self.gridplot_psgens(psgens, [f"rc = {rc:.2f}, qc = {qc:.2f} for l: {l}"
                                        for rc, qc in rq_prod()])
            col.append(grid)

            self.rc_qcut_out_area.objects = col.objects

    #@depends_on_btn_click('execute_btn')
    def on_execute_btn(self, event) -> None:
        """
        Build a new generator from the input file, run it and update out_area.
        """
        with ButtonContext(event.obj), Loading(self.out_area):
            oncv_input = self.get_oncv_input()
            psgen = OncvGenerator(input_str=str(oncv_input), calc_type=self.calc_type)
            print("Running in workdir:", psgen.workdir)
            psgen.start()
            retcode = psgen.wait()
            #psget.check_status()

            #if psgen.status != psgen.S_OK:
            #    cprint("oncvpsp returned %s. Exiting" % retcode, "red")
            #    return 1

            #out_path = self._execute_stdout_path = psgen.stdout_path

            # Parse the output file
            #onc_parser = OncvParser(out_path)
            #onc_parser.scan()
            #if not onc_parser.run_completed:
            #    cprint("oncvpsp output is not complete. Exiting", "red")
            #    return 1

            ## Build the plotter and add figures to out_area.
            #plotter = onc_parser.get_plotter()

            # TODO:
            # Tranfer final output file.

            self._update_out_area(psgen, oncv_input)

    def _update_out_area(self, psgen, oncv_input: OncvInput) -> None:
        with Loading(self.out_area):
            #self.psgen_to_save = psgen
            plotter = psgen.parser.get_plotter()

            if plotter is None:
                self.out_area.objects = pn.Column("## Plotter is None")
                return

            if self.plotlyFlag:
                _m = functools.partial(ply, with_divider=False)

            else:
                _m = functools.partial(mpl, with_divider=False, dpi=self.dpi)

            save_btn = pnw.Button(name="Save output", button_type='primary')
            save_btn.on_click(self.on_save_btn)

            new_rows = [
                pn.Row(save_btn, self.in_filepath),
                pn.layout.Divider(),
                "## Pseudized Wavefunctions",
                pn.Row(_m(plotter.plot_radial_wfs(show=False)),
                    self.get_rc_widgets(oncv_input), height=600),
                pn.Row(_m(plotter.plot_radial_wfs(what="scattering_states", show=False)), height=600),
                pn.layout.Divider(),
                "## Logder and convergence profile",
                pn.Row(_m(plotter.plot_atanlogder_econv(show=False)),
                    self.get_qcut_widgets(oncv_input), height=500),
                pn.Row(pn.Spacer(), self.get_debl_widgets(oncv_input), align='end', height=300),
                pn.layout.Divider(),
                "## Pseudized local part",
                pn.Row(_m(plotter.plot_potentials(show=False)),
                    self.get_rc5_widgets(oncv_input), height=400),
                pn.Row(pn.Spacer(), self.get_dvloc0_widgets(oncv_input), align="end", height=300),
                pn.layout.Divider(),
                "## Model core charge",
                pn.Row(_m(plotter.plot_densities(show=False)),
                    self.get_rhomodel_widgets(oncv_input), height=600),
                pn.Row(_m(plotter.plot_den_formfact(show=False)), height=600),
                pn.layout.Divider(),
                "## Projectors",
                pn.Row(_m(plotter.plot_projectors(show=False)), height=600),
                pn.layout.Divider(),
            ]

            self.out_area.objects = new_rows
            #self.tabs[0].active = 1

    def on_save_btn(self, event) -> None:
        with ButtonContext(event.obj):
            print("on_save_button")
            #self._execute_stdout_path
            out_path = self.in_filepath.value
            if not out_path:
                raise ValueError("out_path cannot be empty.")

            #self.psgen_to_save = psgen

            # TODO: Transfer final output file.
            #shutil.copy(psgen.stdout_path, out_path)



#!/usr/bin/env python
r"""
MgB2 Fermi surface
==================

This example shows how to plot the Fermi surface with matplotlib
"""
from abipy import abilab
import abipy.data as abidata

with abilab.abiopen(abidata.ref_file("mgb2_kmesh181818_FATBANDS.nc")) as fbnc_kmesh:
    ebands = fbnc_kmesh.ebands

# Build ebands in full BZ.
eb3d = ebands.get_ebands3d()

# Use matplotlib to plot isosurfaces corresponding to the Fermi level (default)
# Warning: requires skimage package, rendering could be slow.
eb3d.plot_isosurfaces()

# Alternatively, it's possible to export the data in xcrysden format
# and then use `xcrysden --bxsf mgb2.bxsf`
#eb3d.to_bxsf("mgb2.bxsf")

# If you have mayavi installed, try:
#eb3d.mvplot_isosurfaces()



#!/usr/bin/env python
"""Benchmark for DDK calculations."""
import sys
import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_inputs(paw=False):
    """
    This function constructs the input files for the DDK in SiC.
    Return: GS input, NSCF input, DDK input.
    """
    pseudos = abidata.pseudos("14si.pspnc", "6c.pspnc") if not paw else \
              abidata.pseudos("Si.GGA_PBE-JTH-paw.xml", "6c.lda.atompaw")

    structure = abidata.structure_from_ucell("SiC")

    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=3)

    # Global variables.
    global_vars = dict(
        ecut=32,
        pawecutdg=32 if paw else None,
        nband=24,
        nbdbuf=4,
        # Definition of the SCF procedure
        nstep=100,
        diemac=9.0,
    )

    multi.set_vars(global_vars)

    # Definition of the k-point grid
    multi.set_kmesh(
        ngkpt=[16, 16, 16],
        #ngkpt=[2, 2, 2],
        shiftk=[0.0, 0.0, 0.5,
                0.0, 0.5, 0.0,
                0.5, 0.0, 0.0,
                0.5, 0.5, 0.5]
    )

    multi[0].set_vars(
      kptopt=1,          # Automatic generation of k points with symmetries.
      tolvrs=1.0e-6,
      nband=12,
      nbdbuf=2,
    )

    multi[1].set_vars(
      iscf=-2,
      #getwfk=1,
      #getden=1,
      kptopt=3,
      tolwfr=1.0e-6,
    )

    multi[2].set_vars(
      #getwfk=2,
      #getden=1,
      iscf=-3,        # Need this non-self-consistent option for d/dk
      rfelfd=2,       # Calculate d/dk wave function only
      tolwfr=1.0e-18, # Use wave function residual criterion instead
      kptopt=3,
      rfdir=[1, 0, 0],
      nstep=1,
      nline=0,
    )

    gs_inp, nscf_inp, ddk_inp = multi.split_datasets()

    return gs_inp, nscf_inp, ddk_inp


def build_flow(options):
    gs_inp, nscf_inp, ddk_inp = make_inputs()

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    ebands_work = flowtk.BandStructureWork(gs_inp, nscf_inp)
    flow.register_work(ebands_work)
    flow.exclude_from_benchmark(ebands_work)

    # Get the list of possible parallel configurations from abinit autoparal.
    max_ncpus, min_eff = options.max_ncpus, options.min_eff
    print("Getting all autoparal confs up to max_ncpus: ",max_ncpus," with efficiency >= ",min_eff)

    pconfs = ddk_inp.abiget_autoparal_pconfs(max_ncpus, autoparal=1)
    if options.verbose: print(pconfs)

    work = flowtk.Work()
    for conf, omp_threads in product(pconfs, options.omp_list):
        mpi_procs = conf.mpi_ncpus
        if not options.accept_conf(conf, omp_threads): continue

        manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
        inp = ddk_inp.new_with_vars(conf.vars)
        work.register_ddk_task(inp, manager=manager, deps={ebands_work[1]: "WFK"})

    print("Found %d configurations" % len(work))
    flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Phonons + dynamical quadrupoles
===============================

This example shows how to compute the dynamical matrix of GaP on a user-defined q-mesh
including Born effective charges, the macroscopic dielectric matric and the dynamical quadrupoles Q*.
The final results (out_DDB, out_DVDB) will be produced automatically at the end of the run
and saved in ``flow_phonons_with_quad/outdata/``.

The Q* tensor may be needed to improve the accuracy of the Fourier interpolation of the phonon frequencies,
especially in the long-wavelength limit |q| --> 0.

This example is based on  <https://docs.abinit.org/tests/tutorespfn/Input/tlw_4.abi>
Note that only selected features are compatible with dynamical quadrupoles.
Please consult <https://docs.abinit.org/topics/longwave/>
"""

import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata

from abipy import flowtk


def build_flow(options):
    """
    Create a `Flow` for phonon calculations including
    eps_inf, Born effective charges and  dynamical quadrupoles
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    pseudos = abidata.pseudos("31-Ga.LDA.fhi", "15-P.LDA.fhi")

    # Initialize the structure from an abinit string
    # Other approaches (e.g. from cif files) are available as well.
    structure = abilab.Structure.from_abistring("""
 acell 1.0 1.0 1.0
 rprim 0.0 5.023016083002 5.023016083002
       5.023016083002 0.0 5.023016083002
       5.023016083002 5.023016083002 0.0

 # Definition of the atom types and positions
 # ******************************************
 ntypat 2
 znucl  31 15
 natom 2
 typat 1 2
 xred
         0.0000000      0.0000000      0.0000000
         0.2500000      0.2500000      0.2500000
""")

    # Build input for GS calculation
    scf_input = abilab.AbinitInput(structure, pseudos=pseudos)

    # We use parameters similar to the ones in https://docs.abinit.org/tests/tutorespfn/Input/tlw_4.abi
    scf_input.set_vars(
        nband=4,
        ecut=5.0,
        ngkpt=[4, 4, 4],
        shiftk=[
            [0.5, 0.5, 0.5],
            [0.5, 0.0, 0.0],
            [0.0, 0.5, 0.0],
            [0.0, 0.0, 0.5],
        ],
        diemac=12.0,
        nstep=100,
        tolvrs=1.0e-10,
        #tolvrs=1.0e-18,  # This is the value used in tlw_4.abi
                          # but it is not always possible to reach this precision in more complex systems.
        #useylm=1,
        #ixc=7
        #iomode=3,
        #paral_kgb=1,
    )

    # At the time of writing, Q* calculations are implemented only for
    # NC scalar-relativistic pseudos without non-linear core correction.
    # This section shows how to use the Pseudo API to perform this kind of check
    # before runnnig the calculation.
    for pseudo in scf_input.pseudos:
        # print(pseudo)
        if not pseudo.isnc:
            raise RuntimeError("Only NC pseudos are compatible with Q*")
        if pseudo.has_nlcc:
            raise RuntimeError("NLCC is not compatible with Q*")

    # Initialize the flow
    flow = flowtk.Flow(workdir=options.workdir, manager=options.manager)

    # Compute phonons on the ddb_ngqpt q-mesh.
    # Include Born effective charges and dynamical quadrupoles via `with_quad=True`.
    #
    #ddb_ngqpt = [1, 1, 1]
    ddb_ngqpt = [2, 2, 2]
    ph_work = flowtk.PhononWork.from_scf_input(scf_input, qpoints=ddb_ngqpt, is_ngqpt=True,
                                               with_becs=True, with_quad=True, ndivsm=5)

    # Add the phonon work to the flow
    flow.register_work(ph_work)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Effective masses with DFPT
==========================

Flow to compute effective masses with DFPT.
Two options are available:

    - EffMassDFPTWork --> Run DFPT calculation directly assuming the location
                          of the band edges is already known.
    - EffMassAutoDFPTWork --> Run NSCF calculation to find band edges, then use DFPT.
"""

import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def make_scf_input(usepaw=0, nspinor=1):
    """Returns input for GS-SCF calculation."""
    if nspinor == 1:
        pseudos = abidata.pseudos("14si.pspnc") if usepaw == 0 else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")
    else:
        pseudos = abidata.pseudos("Si_r.psp8") if usepaw == 0 else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")

    # See https://docs.abinit.org/tests/v7/Input/t82.in
    structure = dict(
         ntypat=1,
         natom=2,
         typat=[1, 1],
         znucl=14,
         #acell=3 * [10.26310667319252],
         acell=3 * [10.2073557], # 5.4015 Ang
         rprim=[[0.0,  0.5,  0.5],
                [0.5,  0.0,  0.5],
                [0.5,  0.5,  0.0]],
         xred=[[0.0 , 0.0 , 0.0],
               [0.25, 0.25, 0.25]],
    )

    scf_input = abilab.AbinitInput(structure=structure, pseudos=pseudos)

    # Global variables
    nband = 8 if nspinor == 1 else 16
    scf_input.set_vars(
        ecut=8,
        nband=nband,
        nspinor=nspinor,
        nstep=100,
        tolvrs=1e-8,
    )

    if scf_input.ispaw:
        scf_input.set_vars(pawecutdg=2 * scf_input["ecut"])

    # Set k-mesh
    scf_input.set_kmesh(ngkpt=[8, 8, 8], shiftk=[0, 0, 0])

    return scf_input


def build_flow(options):
    # Set working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Get the SCF input (without SOC)
    scf_input = make_scf_input(nspinor=1, usepaw=1)

    # Build the flow.
    from abipy.flowtk.effmass_works import EffMassDFPTWork, EffMassAutoDFPTWork
    flow = flowtk.Flow(workdir=options.workdir, manager=options.manager)

    # Compute effective masses for each k in k0_list.
    # effmass_bands_f90 defines the band range for each k in k0_list
    # Here we are interested in the effective masses at the Gamma point for the valence bands
    effmass_bands_f90 = [1, 4] if scf_input["nspinor"] == 1 else [1, 8]
    work = EffMassDFPTWork.from_scf_input(scf_input, k0_list=(0, 0, 0),
                                          effmass_bands_f90=effmass_bands_f90)
    flow.register_work(work)

    # or use this Work to detect band edges automatically but increase ndivsm and decrease tolwfr!
    # you may want to use a negative value of ndivsm (e.g. -20) to use the pymatgen density_line
    # convention. This is useful to avoid problems with high-symmetry
    # k-paths containing very small segments.
    # In this case, indeed,  ndivsm > 0 (Abinit variable) can easily generate thousands of k-points.
    work = EffMassAutoDFPTWork.from_scf_input(scf_input, ndivsm=-5, tolwfr=1e-12)
    flow.register_work(work)

    return flow


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_effmass_dfpt -s
#



#!/usr/bin/env python
"""
This script opens one of the output files produced by Abinit (usually in netcdf format but
other files are supported as well). By default the script starts an interactive ipython
session so that one can interact with the file and call its methods.
Alternatively, it is possible to generate automatically a jupyter notebook to execute code.
"""
from __future__ import annotations

import sys
import os
import argparse
import subprocess
import abipy.tools.cli_parsers as cli
from abipy.tools.plotting import Exposer

from pprint import pprint
from shutil import which
from monty.termcolor import cprint
from monty.functools import prof_main
from abipy import abilab


def make_and_open_notebook(options):
    """
    Generate a jupyter notebook and open it in the browser.
    Return system exit code.

    Raise:
        RuntimeError if jupyter is not in $PATH
    """
    import os
    import nbformat
    nbf = nbformat.v4
    nb = nbf.new_notebook()

    nb.cells.extend([
        nbf.new_markdown_cell("## This is an auto-generated notebook for %s" % os.path.relpath(options.filepath)),
        nbf.new_code_cell("""\

%matplotlib notebook
import numpy as np
#import seaborn as sns
#sns.set(context='notebook', style='darkgrid', palette='deep',
#        font='sans-serif', font_scale=1, color_codes=False, rc=None)
from abipy import abilab

"""),
        nbf.new_code_cell("abifile = abilab.abiopen('%s')" % options.filepath)
    ])

    import io, tempfile
    _, nbpath = tempfile.mkstemp(prefix="abinb_", suffix='.ipynb', dir=os.getcwd(), text=True)

    with io.open(nbpath, 'wt', encoding="utf8") as f:
        nbformat.write(nb, f)

    if which("jupyter") is None:
        raise RuntimeError("Cannot find jupyter in PATH. Install it with `pip install`")

    if not options.classic_notebook:
        # Use jupyter-lab instead of classic notebook
        has_jupyterlab = which("jupyter-lab") is not None
        appname = "jupyter-lab" if has_jupyterlab else "jupyter notebook"
    else:
        appname = "jupyter notebook"

    if options.foreground:
        return os.system("%s %s" % (appname, nbpath))
    else:
        fd, tmpname = tempfile.mkstemp(text=True)
        print(tmpname)
        cmd = "%s %s" % (appname, nbpath)
        print("Executing:", cmd, "\nstdout and stderr redirected to %s" % tmpname)
        process = subprocess.Popen(cmd.split(), shell=False, stdout=fd, stderr=fd)
        cprint("pid: %s" % str(process.pid), "yellow")
        return 0


def get_epilog() -> str:
    s = """\
======================================================================================================
Usage example:

    abiopen.py FILE          => Open file in ipython shell.
    abiopen.py FILE -p       => Print info on object to terminal.
    abiopen.py FILE -e       => Generate matplotlib figures automatically.
                                Use -sns to activate seaborn settings.
    abiopen.py FILE -eweb    => Generate matplotlib figures, show them in the $BROWSER.
    abiopen.py FILE -ply     => Generate plotly figures automatically. Show them in the $BROWSER.
                                Note that not all FILEs support plotly.
    abiopen.py FILE -pn      => Generate GUI in web BROWSER to interact with FILE. Requires panel package.
    abiopen.py FILE -nb      => Generate jupyter-lab notebook.
    abiopen.py FILE -cnb     => Generate classic jupyter notebook.

where `FILE` is any file supported by abipy/pymatgen e.g. Netcdf files, Abinit input, POSCAR, xsf.
File extensions supported (including zipped files with extension in ".bz2", ".gz", ".z"):
Use `-v` to increase verbosity level (can be supplied multiple times e.g -vv).

JSON file are supported as well. In this case, abiopen.py tries to reconstruct python objects
assuming JSON document in MSONable format and then invokes ipython with the `data` object.
Use `-e` or `--notebook` or `--panel` to print the JSON dictionary without reconstructing python objects.
======================================================================================================

Table mapping file extension to AbiPy object:

"""
    return s + abilab.abiopen_ext2class_table()


def get_parser(with_epilog=False):
    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--loglevel', default="ERROR", type=str,
        help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")
    parser.add_argument('-V', '--version', action='version', version=abilab.__version__)

    parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
        help='verbose, can be supplied multiple times to increase verbosity')

    parser.add_argument("filepath", help="File to open. See table below for the list of supported extensions.")

    # notebook options.
    parser.add_argument('-nb', '--notebook', action='store_true', default=False, help="Open file in jupyter notebook")
    parser.add_argument('--classic-notebook', "-cnb", action='store_true', default=False,
                        help="Use classic jupyter notebook instead of jupyterlab.")
    parser.add_argument('--no-browser', action='store_true', default=False,
                        help=("Start the jupyter server to serve the notebook "
                              "but don't open the notebook in the browser.\n"
                              "Use this option to connect remotely from localhost to the machine running the kernel"))
    parser.add_argument('--foreground', action='store_true', default=False,
                        help="Run jupyter notebook in the foreground.")

    # print option
    parser.add_argument('-p', '--print', action='store_true', default=False, help="Print python object and return.")

    # panel options
    parser.add_argument("-pn", '--panel', action='store_true', default=False,
                        help="Open Dashboard in web browser, requires panel package.")
    parser.add_argument("-pnt", "--panel-template", default="FastList", type=str,
                        help="Specify template for panel dasboard." +
                             "Possible values are: FastList, FastGrid, Golden, Bootstrap, Material, React, Vanilla." +
                             "Default: FastList"
                        )
    parser.add_argument("--port", default=0, type=int, help="Allows specifying a specific port when serving panel app.")
    #add_expose_options_to_parser(parser)

    # Expose option.
    parser.add_argument('-e', '--expose', action='store_true', default=False,
        help="Open file and generate matplotlib figures automatically by calling expose method.")
    parser.add_argument("-s", "--slide-mode", default=False, action="store_true",
        help="Iterate over figures. Expose all figures at once if not given on the CLI.")
    parser.add_argument("-t", "--slide-timeout", type=int, default=None,
        help="Close figure after slide-timeout seconds (only if slide-mode). Block if not specified.")
    parser.add_argument('-sns', "--seaborn", const="paper", default=None, action='store', nargs='?', type=str,
            help='Use seaborn settings. Accept value defining context in ("paper", "notebook", "talk", "poster"). Default: paper')
    parser.add_argument('-mpl', "--mpl-backend", default=None,
        help=("Set matplotlib interactive backend. "
              "Possible values: GTKAgg, GTK3Agg, GTK, GTKCairo, GTK3Cairo, WXAgg, WX, TkAgg, Qt4Agg, Qt5Agg, macosx."
              "See also: https://matplotlib.org/faq/usage_faq.html#what-is-a-backend."))
    parser.add_argument("-ew", "--expose-web", default=False, action="store_true",
            help='Generate matplotlib plots in $BROWSER instead of X-server. WARNING: Not all the features are supported.')
    parser.add_argument("-ply", "--plotly", default=False, action="store_true",
            help='Generate plotly plots in $BROWSER instead of matplotlib. WARNING: Not all the features are supported.')

    return parser


def serve_kwargs_from_options(options):

    if options.no_browser:
        print("""
Use:

    ssh -N -f -L localhost:{port}:localhost:{port} username@your_remote_cluster

for port forwarding.
""")

    import abipy.panels as mod
    assets_path = os.path.join(os.path.dirname(mod.__file__), "assets")

    return dict(
        debug=options.verbose > 0,
        show=not options.no_browser,
        port=options.port,
        static_dirs={"/assets": assets_path},
        #address=address,
        #websocket_origin="{address}:{port}",
    )



@prof_main
def main():
    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg:
            sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse the command line.
    try:
        options = parser.parse_args()
    except Exception:
        show_examples_and_exit(error_code=1)

    cli.set_loglevel(options.loglevel)

    ##############################################################################################
    # Handle meta options i.e. options that set other options.
    # OK, it's not very clean but I haven't find any parse API to express this kind of dependency.
    ##############################################################################################
    if options.plotly: options.expose = True
    if options.expose_web: options.expose = True
    if options.classic_notebook: options.notebook = True

    if options.verbose > 2: print(options)

    # Set matplotlib backend
    if options.mpl_backend is not None:
        import matplotlib
        matplotlib.use(options.mpl_backend)

    # Use seaborn settings.
    if options.seaborn:
        import seaborn as sns
        sns.set(context=options.seaborn, style='darkgrid', palette='deep',
                font='sans-serif', font_scale=1, color_codes=False, rc=None)

    if not os.path.exists(options.filepath):
        raise RuntimeError("%s: no such file" % options.filepath)

    if options.filepath.endswith(".json"):
        return handle_json(options)

    if options.filepath.endswith(".csv"):
        return handle_csv(options)

    if options.filepath.endswith(".traj"):
        return handle_ase_traj(options)

    if os.path.basename(options.filepath) == "flows.db":
        from abipy.flowtk.launcher import print_flowsdb_file
        return print_flowsdb_file(options.filepath)

    if not options.notebook:
        abifile = abilab.abiopen(options.filepath)
        if options.print:
            # Print object to terminal.
            if hasattr(abifile, "to_string"):
                #print(f"Calling {abifile.__class__}.to_string with verbose: {verbose}")
                print(abifile.to_string(verbose=options.verbose))
            else:
                print(abifile)
            return 0

        elif options.expose:
            # Print info to terminal
            if hasattr(abifile, "to_string"):
                print(abifile.to_string(verbose=options.verbose))
            else:
                print(abifile)

            # Generate plots automatically.
            if options.plotly:
                # plotly version
                if hasattr(abifile, "plotly_expose"):
                    abifile.plotly_expose(verbose=options.verbose)
                else:
                    cprint("`%s` does not implement plotly_expose method" % type(abifile), "red")

            elif hasattr(abifile, "expose"):
                # matplotlib version
                abifile.expose(slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                               use_web=options.expose_web, verbose=options.verbose)
            else:
                if not hasattr(abifile, "yield_figs"):
                    raise TypeError("Object of type `%s` does not implement (expose or yield_figs methods" % type(abifile))
                from abipy.tools.plotting import MplExposer
                with MplExposer(slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                                verbose=options.verbose) as e:
                    e(abifile.yield_figs())

            return 0

        elif options.panel:
            if not hasattr(abifile, "get_panel"):
                raise TypeError("Object of type `%s` does not implement get_panel method" % type(abifile))

            import matplotlib
            matplotlib.use("Agg")
            pn = abilab.abipanel()

            app = abifile.get_panel(template=options.panel_template)
            serve_kwargs = serve_kwargs_from_options(options)

            return pn.serve(app, **serve_kwargs)

        # Start ipython shell with namespace
        # Use embed because I don't know how to show a header with start_ipython.
        import IPython
        IPython.embed(header="""
The Abinit file object is associated to the `abifile` python variable.
Use `abifile.<TAB>` to list available methods.
Use e.g. `abifile.plot?` to access docstring and `abifile.plot??` to visualize source.
Use `print(abifile)` to print the object.
""")

    else:
        # Call specialized method if the object is a NotebookWriter
        # else generate simple notebook by calling `make_and_open_notebook`
        cls = abilab.abifile_subclass_from_filename(options.filepath)
        if hasattr(cls, "make_and_open_notebook"):
            if hasattr(cls, "__exit__"):
                with abilab.abiopen(options.filepath) as abifile:
                    return abifile.make_and_open_notebook(foreground=options.foreground,
                                                          classic_notebook=options.classic_notebook,
                                                          no_browser=options.no_browser)
            else:
                abifile = abilab.abiopen(options.filepath)
                return abifile.make_and_open_notebook(foreground=options.foreground,
                                                      classic_notebook=options.classic_notebook,
                                                      no_browser=options.no_browser)
        else:
            return make_and_open_notebook(options)

    return 0


def handle_ase_traj(options):
    """Handle ASE trajectory file."""
    from abipy.ml.aseml import AseTrajectoryPlotter
    plotter = AseTrajectoryPlotter.from_file(options.filepath)

    print(plotter.to_string(verbose=options.verbose))
    if options.expose:
        if len(plotter.traj) > 1:
            plot_kws = dict(show=False)
            with Exposer.as_exposer("mpl") as e:
                e(plotter.plot(**plot_kws))
                e(plotter.plot_lattice(**plot_kws))

    return 0


def handle_csv(options):
    """Handle CSV file."""
    df = pd.read_csv(options.filepath)

    def print_df():
        print("=== Dataframe info ===")
        print(df.info())
        print("=== Dataframe describe ===")
        print(df.describe())

    if options.notebook:
        raise NotImplementedError("")
        # Visualize JSON document in jupyter
        #cmd = "jupyter-lab %s" % options.filepath
        #print("Executing:", cmd)
        #process = subprocess.Popen(cmd.split(), shell=False) #, stdout=fd, stderr=fd)
        #cprint("pid: %s" % str(process.pid), "yellow")
        return 0

    elif options.panel:
        raise NotImplementedError("")
        # Visualize JSON document in panel dashboard.
        #pn = abilab.abipanel()
        #with abilab.abiopen(options.filepath) as json_file:
        #    app = json_file.get_panel()

        #serve_kwargs = serve_kwargs_from_options(options)
        #return pn.serve(app, **serve_kwargs)

    else:
        if options.print:
            # Print python object to terminal.
            print_df()
            return 0
        elif options.expose:
            print_df()
            raise NotImplementedError("")
            return 0

        # Start ipython shell with namespace
        # Use embed because I don't know how to show a header with start_ipython.
        print_df()
        import IPython
        IPython.embed(header="""
The pandas DataFrame initialized from the csv file can be accesssed via the `df` python variable.
""")

    return 0


def handle_json(options):
    """Handle JSON file."""
    if options.notebook:
        # Visualize JSON document in jupyter
        cmd = "jupyter-lab %s" % options.filepath
        print("Executing:", cmd)
        process = subprocess.Popen(cmd.split(), shell=False) #, stdout=fd, stderr=fd)
        cprint("pid: %s" % str(process.pid), "yellow")
        return 0

    elif options.panel:
        # Visualize JSON document in panel dashboard.
        pn = abilab.abipanel()
        with abilab.abiopen(options.filepath) as json_file:
            app = json_file.get_panel()

        serve_kwargs = serve_kwargs_from_options(options)
        return pn.serve(app, **serve_kwargs)

    else:
        if options.print:
            # Print python object to terminal.
            data = abilab.mjson_load(options.filepath)
            pprint(data, indent=4)
            return 0
        elif options.expose:
            # Pretty-print dict to terminal.
            import json
            with open(options.filepath, "rt") as fh:
                data = json.load(fh)
            pprint(data, indent=4)
            return 0

        data = abilab.mjson_load(options.filepath)
        # Start ipython shell with namespace
        # Use embed because I don't know how to show a header with start_ipython.
        import IPython
        IPython.embed(header="""
The object initialized from JSON (MSONable) is associated to the `data` python variable.
""")

    return 0


def handle_flowsdb_file(options):
    """Handle flows.db file."""
    import pandas as pd
    import sqlite3
    with sqlite3.connect(options.filepath) as con:
        df = pd.read_sql_query("SELECT * FROM flows", con)
        abilab.print_dataframe(df, title=options.filepath)
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
e-bands scatter3d
=================

This example shows how to plot a particular "band" in 3D
with matplotlib scatter3D.
The color of the point gives the energy of the state wrt to the Fermi level.
"""
from abipy import abilab
import abipy.data as abidata

# Open the GSR file and extract the band structure.
with abilab.abiopen(abidata.ref_file("ni_kpath_GSR.nc")) as ncfile:
    ncfile.ebands.plot_scatter3d(band=9, spin=0)



#!/usr/bin/env python
r"""
Phonon-limited mobilities in semiconductors
===========================================

This flow computes the phonon-limited mobility in AlAs
using different dense k/q meshes.
"""
import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.core.abinit_units as abu


def build_flow(options):

    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Initialize the flow
    flow = flowtk.Flow(workdir=options.workdir, manager=options.manager)

    # Initialize the tmesh, sigma_kerange, sigma_erange and dense k-meshes for the eph integrations
    tmesh = [300, 300, 1]
    sigma_kerange = [0, 0.5 * abu.eV_Ha]  # 0.5 eV range for the WFK of electrons
    sigma_erange = [0, 0.25 * abu.eV_Ha]  # 0.25 eV range for the EPH computation for electrons
    dense_kmeshes = [
        [30, 30, 30],
        [40, 40, 40],
    ]

    # Initialize the structure and pseudos
    structure = abidata.structure_from_ucell("AlAs")
    pseudos = abidata.pseudos("13al.981214.fhi", "33as.pspnc")
    ecut = 2
    nband = 8

    # Ground-state computation for 1) the phonons and 2) the WFK generation
    scf_input = abilab.AbinitInput(structure, pseudos)

    scf_input.set_vars(
        nband=nband,
        ecut=ecut,
        ngkpt=[4, 4, 4],
        shiftk=[0, 0, 0],
        tolvrs=1.0e-8,
        diemac=9.0,
        #prtden=1,
        #iomode=3,
    )

    # Initialize the first work and add the ground-state task
    work0 = flowtk.Work()
    work0.register_scf_task(scf_input)

    # Band structure calculation to make sure everything is OK
    # Also allows to compare the results obtained with abitk to
    # check the SKW interpolation works properly.
    bs_input = scf_input.make_ebands_input(tolwfr=1e-12, ndivsm=10, nb_extra=5)
    bs_input.set_vars(nstep=100, nbdbuf=1)
    work0.register_nscf_task(bs_input, deps={work0[0]: "DEN"})

    # NSCF input for the WFK needed to interpolate with kerange
    nscf_input = abilab.AbinitInput(structure, pseudos)
    nscf_input.set_vars(
        ecut=ecut,
        nband=nband,
        iscf=-2,
        tolwfr=1e-20,
        prtwf=1,
        ngkpt=[16, 16, 16], # Should be dense enough so that the kerange interpolation works
        shiftk=[0.0, 0.0, 0.0],
    )

    work0.register_nscf_task(nscf_input, deps={work0[0]: "DEN"})
    flow.register_work(work0)

    # Add the phonon work to the flow
    # NB: with_quad is set to False because non-linear core correction is not yet supported.
    ddb_ngqpt = [4, 4, 4]
    ph_work = flowtk.PhononWork.from_scf_task(work0[0], qpoints=ddb_ngqpt,
                                              is_ngqpt=True, with_becs=True, with_quad=False)
    flow.register_work(ph_work)

    # We loop over the dense k-meshes
    for i, sigma_ngkpt in enumerate(dense_kmeshes):

        # Use the kerange trick to generate a WFK file
        multi = nscf_input.make_wfk_kerange_inputs(sigma_kerange=sigma_kerange,
                                                   sigma_ngkpt=sigma_ngkpt)
        kerange_input, wfk_input = multi.split_datasets()

        work_eph = flowtk.Work()
        work_eph.register_kerange_task(kerange_input, deps={work0[2]: "WFK"})
        work_eph.register_nscf_task(wfk_input,
                                    deps={work0[0]: "DEN", work_eph[0]: "KERANGE.nc"})

        # Generate the input file for transport calculations.
        # Use ibte_prep = 1 to activate the iterative BTE.
        eph_input = wfk_input.make_eph_transport_input(ddb_ngqpt=ddb_ngqpt,
                                                       sigma_erange=sigma_erange,
                                                       tmesh=tmesh,
                                                       eph_ngqpt_fine=sigma_ngkpt,
                                                       ibte_prep=1)

        # We compute the phonon dispersion in the EPH code to be able to check they are ok.
        if i == 0:
            eph_input.set_qpath(20)

        work_eph.register_eph_task(eph_input,
                                   deps={work_eph[1]: "WFK", ph_work: ["DDB", "DVDB"]})

        flow.register_work(work_eph)

    flow.allocate(use_smartio=True)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Electron bands box plot
=======================

This example shows how to use seaborn or plotly to draw a box plot
showing the distributions of eigenvalues with respect to the band index.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# Open the file with energies computed with a homogeneous sampling
# of the BZ and extract the band structure.
with abiopen(abidata.ref_file("si_scf_GSR.nc")) as gsr:
    ebands = gsr.ebands

#%%
# Use `swarm=True` to show the datapoints on top of the boxes
ebands.boxplot(swarm=True)

#%%
# For the plotly version, use:
ebands.boxplotly()



# coding: utf-8
"""
Work for computing the Grüneisen parameters with finite differences of DFPT phonons.

WARNING: This code must be tested more carefully.
"""
from __future__ import annotations

import numpy as np

from abipy.core.structure import Structure
from .works import Work, PhononWork


class GruneisenWork(Work):
    """
    This work computes the Grüneisen parameters (derivative of frequencies wrt volume)
    using finite differences and the phonons obtained with the DFPT part of Abinit.
    The Anaddb input file needed to compute Grüneisen parameters will be generated
    in the outdata directory of the flow.

    It is necessary to run three DFPT phonon calculations.
    One is calculated at the equilibrium volume and the remaining two are calculated
    at the slightly larger volume and smaller volume than the equilibrium volume.
    The unitcells at these volumes have to be fully relaxed under the constraint of each volume.
    This Work automates the entire procedure starting from an input for GS calculations.
    """

    @classmethod
    def from_gs_input(cls, gs_inp, voldelta, ngqpt, 
                      tolerance=None, with_becs=False, 
                      ddk_tolerance=None, workdir=None, manager=None) -> GruneisenWork:
        """
        Build the work from an |AbinitInput| representing a GS calculations.

        Args:
            gs_inp: |AbinitInput| representing a GS calculation in the initial unit cell.
            voldelta: Absolute increment for unit cell volume. The three volumes are:
                [v0 - voldelta, v0, v0 + voldelta] where v0 is taken from gs_inp.structure.
            ngqpt: three integers defining the q-mesh for phonon calculations.
            tolerance: dict {"varname": value} with the tolerance to be used in the phonon run.
                None to use AbiPy default.
            with_becs: Activate calculation of Electric field and Born effective charges.
            ddk_tolerance: dict {"varname": value} with the tolerance used in the DDK run if with_becs.
                None to use AbiPy default.
        """
        new = cls(workdir=workdir, manager=manager)
        new.ngqpt = np.reshape(ngqpt, (3,))
        new.with_becs = with_becs
        new.ddk_tolerance = ddk_tolerance
        new.tolerance = tolerance

        if any(gs_inp["ngkpt"] % new.ngqpt != 0):
            raise ValueError("Kmesh and Qmesh must be commensurate.\nGot ngkpt: `%s`\nand ngqpt: `%s`" % (
                             str(gs_inp["ngkpt"]), str(new.ngqpt)))

        # Build three tasks for structural optimization at constant volume.
        v0 = gs_inp.structure.volume
        if voldelta <= 0:
            raise ValueError("voldelta must be > 0 but got %s" % voldelta)
        volumes = [v0 - voldelta, v0, v0 + voldelta]
        if any(v <= 0 for v in volumes):
            raise ValueError("volumes must be > 0 but got %s" % str(volumes))

        # Keep a copy of the GS input that will be used to generate the Phonon Works
        new.gs_inp = gs_inp.deepcopy()

        new.relax_tasks = []
        for vol in volumes:
            # Build new structure
            new_lattice = gs_inp.structure.lattice.scale(vol)
            new_structure = Structure(new_lattice, gs_inp.structure.species, gs_inp.structure.frac_coords)
            new_input = gs_inp.new_with_structure(new_structure)
            # Set variables for structural optimization at constant volume.
            new_input.pop_tolerances()
            new_input.set_vars(optcell=3, ionmov=3, tolvrs=1e-10, toldff=1.e-6)
            new_input.set_vars_ifnotin(ecutsm=0.5, dilatmx=1.05)
            t = new.register_relax_task(new_input)
            new.relax_tasks.append(t)

        return new

    def on_all_ok(self):
        """
        This method is called once the `Work` is completed i.e. when all the tasks have reached status S_OK.
        """
        self.add_phonon_works_and_build()
        return super().on_all_ok()

    def add_phonon_works_and_build(self) -> None:
        """
        Get relaxed structures from the tasks, build Phonons works with new structures.
        Add works to the flow and build new directories.
        """
        ddb_paths, struct_middle, middle_idx = [], None, None
        relaxed_structs = []

        for i, task in enumerate(self.relax_tasks):
            relaxed_structure = task.get_final_structure()
            relaxed_structs.append(relaxed_structure)
            if i == len(self.relax_tasks) // 2:
                middle_idx, struct_middle = i, relaxed_structure

            # work to compute phonons with new structure.
            gsinp_vol = self.gs_inp.new_with_structure(relaxed_structure)
            work = PhononWork.from_scf_input(gsinp_vol, self.ngqpt, is_ngqpt=True, tolerance=self.tolerance,
                                             with_becs=self.with_becs, ddk_tolerance=self.ddk_tolerance)
            # Add it to the flow.
            self.flow.register_work(work)

            ddb_paths.append(work.outdir.path_in("out_DDB"))

        # Write Anaddb input file to compute Gruneisen parameters in flow.outdata.
        from abipy.abio.inputs import AnaddbInput
        anaddb_inp = AnaddbInput.phbands_and_dos(struct_middle, self.ngqpt, nqsmall=20, ndivsm=20,
                                                 chneut=1 if self.with_becs else 0,
                                                 dipdip=1 if self.with_becs else 0,
                                                 lo_to_splitting=self.with_becs,
                                                 comment="Anaddb input file for Grunesein parameters")
        # Add DDB files for Gruns
        anaddb_inp["gruns_nddbs"] = len(ddb_paths)
        anaddb_inp["gruns_ddbs"] = "\n" + "\n".join('"%s"' % p for p in ddb_paths)
        in_path = self.flow.outdir.path_in("anaddb_gruns.in")
        anaddb_inp.write(in_path)

        files_file = []
        app = files_file.append
        app(in_path)                                        # 1) Path of the input file
        app(self.flow.outdir.path_in("anaddb_gruns.out"))   # 2) Path of the output file
        app(ddb_paths[middle_idx])                          # 3) Input derivative database (not used if Gruns)
        for i in range(4):
            app("FOOBAR")

        with open(self.flow.outdir.path_in("anaddb_gruns.files"), "wt") as fh:
            fh.write("\n".join(files_file))

        #task = AbinitTask.temp_shell_task(anaddb_inp, workdir=work.outdir, manager=self.manager)
        #task.start_and_wait(autoparal=False)

        #with_ebands = False
        #if with_ebands:
        #    bands_work = Work(manager=self.manager)
        #    for i, structure in enumerate(relaxed_structs):
        #        nscf_inp = self.gs_inp.new_with_structure(structure)
        #        nscf_inp.pop_tolerances()
        #        nscf_inp.set_kpath(ndivsm, kptbounds=None, iscf=-2)
        #        nscf_inp["tolwfr"] = 1e-18
        #        work.register_nscf_task(nscf_inp, deps={self.relax_tasks[i]: "DEN"})

        #    # Add it to the flow.
        #    self.flow.register_work(work)

        # Allocate new works and update the pickle database.
        self.flow.allocate()
        self.flow.build_and_pickle_dump()




from phonopy import Phonopy
import numpy as np

from pymatgen.io.phonopy import get_pmg_structure,get_phonopy_structure

from abipy.core.abinit_units import eV_to_THz
from abipy.dfpt.converters import phonopy_to_abinit
import os,shutil
from abipy.embedding.utils_ifc import accoustic_sum,map_two_structures_coords,clean_structure
from abipy.core.structure import Structure


class Embedded_phonons(Phonopy):
    """
    Defines a Phonopy object implementing the Interatomic Force Constants embedding method defined in:
    https://pubs.acs.org/doi/10.1021/acs.chemmater.3c00537    
    https://iopscience.iop.org/article/10.1088/1367-2630/16/7/073026/meta
    https://journals.aps.org/prmaterials/abstract/10.1103/PhysRevMaterials.5.084603

    """
    
    # Inherits from Phonopy class
    def __init__(self,stru_pristine,stru_defect,stru_emb,ifc_emb,nac_params) :
        """
        :param stru_pristine: Supercell pristine structure
        :param stru_defect: Supercell defect structure
        :param stru_emb: Supercell embedded structure
        :param ifc_emb: Interatomic force constant associated to the supercell embedded structure
        :param nac_params : Non-analytical parameters associated to the supercell embedded structure, in phonopy format. 
        """
        super().__init__(unitcell=get_phonopy_structure(stru_emb),
                       supercell_matrix=[1,1,1],
                       primitive_matrix=np.eye(3),)
        self.force_constants=ifc_emb
        self.nac_params=nac_params
        self.stru_pristine=stru_pristine
        self.stru_defect=stru_defect
        self.stru_emb=stru_emb



    @classmethod
    def from_phonopy_instances(cls,
                               phonopy_pristine,
                               phonopy_defect,
                               structure_defect_wo_relax,
                               main_defect_coords_in_pristine,
                               main_defect_coords_in_defect, 
                               substitutions_list:list=None,  #index in pristine,specie ex: [0,"Eu"]
                               vacancies_list:list=None,       # index in pristine ex: [13,14]
                               interstitial_list:list=None,  # species, cart_coord ex: [['Eu',[0,0,0]],['Ce','[0,0,3]']]
                               tol_mapping=0.01,
                               cut_off_mode='auto',rc_1=None,rc_2=None,
                               factor_ifc=1,
                               verbose=0,
                               asr=True) -> Phonopy :
        """

        Args:
            phonopy_pristine: Phonopy object of the pristine structure
            phonopy_defect  : Phonopy object of the defect structure
            structure_defect_wo_relax : Supercell structure associated to the defect structure, but without relaxation. Needed for an easier mapping.
                Should corresponds to order of phonopy_defect structure!
            main_defect_coords_in_pristine : Main coordinates of the defect in pristine structure, if defect complex, can be set to the
                center of mass of the complex
            main_defect_coords_in_defect : Main coordinates of the defect in defect structure, if defect complex, can be set to the
                center of mass of the complex
            substitutions_list : List of substitutions infos [index,specie], ex : [[0, "Eu"],[1,"N"]]
            vacancies_list : List of indices where the vacancies are, ex: [13,14]
            interstitial_list : List of interstitial infos [specie, cart_coord], ex [['Eu',[0,0,0]],['Ce','[0,0,3]']]
            tol_mapping : Tolerance in angstrom for the mapping between structures
            cut_off_mode : Cut off mode for the radii of the sphere centered around the defect (rc_2). if 'auto' : the code tries to find the largest sphere 
                inscribed in the defect supercell. If 'manual' :  rc_1 and rc_2 should be provided.
            rc_1 : Radii of the sphere centered around the defect outside which the IFCs are set to zero, allows to get sparse matrix. 
            rc_2 : Radii of the sphere centered around the defect where the IFCs of the defect computation are included 
            factor_ifc : Multiply the IFCs inside the sphere of radii rc_2 by factor_ifc, usefull to introduce fictious high-frequency local mode 
            verbose : Print explicitely all the IFCs replacements 
            asr : If True, re-enforce acoustic sum rule after IFCs embedding, following eq. (S4) of https://pubs.acs.org/doi/10.1021/acs.chemmater.3c00537

        Returns:
            A new phonopy object with the embedded structure and embedded IFCs.

        """
        
        
        ########################
        # Structures manipulations and mapping
        ########################

        stru_defect=get_pmg_structure(phonopy_defect.supercell)
        stru_pristine=get_pmg_structure(phonopy_pristine.supercell)

        stru_emb=stru_pristine.copy()


        if vacancies_list is not None:
            stru_emb.remove_sites(indices=vacancies_list)

        if substitutions_list is not None:
            for sub in substitutions_list: 
                stru_emb.replace(sub[0],sub[1])

        if interstitial_list is not None:
            for inter in interstitial_list: 
                stru_emb.append(species=inter[0],coords=inter[1],coords_are_cartesian=True)

        stru_emb=clean_structure(stru_emb,defect_coord=main_defect_coords_in_pristine)
        structure_defect_wo_relax=clean_structure(structure_defect_wo_relax,defect_coord=main_defect_coords_in_defect)
        

        mapping=map_two_structures_coords(structure_defect_wo_relax,stru_emb,tol=tol_mapping)

        ########################
        # Init of the IFCs
        ########################

        ifc_pristine=phonopy_pristine.force_constants
        ifc_defect=phonopy_defect.force_constants

        # in case of IFC with vacancy, remove ifcs of the vac site
        if vacancies_list is not None:
            ifc_pristine=np.delete(ifc_pristine,vacancies_list,0)

            ifc_pristine=np.delete(ifc_pristine,vacancies_list,1)

        # interstitial
        if interstitial_list is not None:
            for inter in interstitial_list:
                ifc_pristine=np.append(ifc_pristine,np.zeros(shape=[1,len(ifc_pristine),3,3]),axis=0)

                ifc_pristine=np.append(ifc_pristine,np.zeros(shape=[len(ifc_pristine),1,3,3]),axis=1)

    

        ########################
        # Print infos
        ########################

        print(f"Number of atoms in the pristine supercell      : {len(stru_pristine)}")
        print(f"Number of atoms in the defective supercell     : {len(stru_defect)}")

        print(f"Defect infos")
        if substitutions_list is not None:
            print(f"    Substitutions :")
            for sub in substitutions_list:
                print(f"       {sub[0]}, {stru_pristine[sub[0]].coords}, {stru_pristine[sub[0]].species} replaced by {sub[1]}")

        if vacancies_list is not None:
            print(f"    Vacancies :" )
            for vac in vacancies_list:
                print(f"       {vac}, {stru_pristine[vac].coords}, {stru_pristine[vac].species} removed")

        if interstitial_list is not None:
            print(f"    Interstitials :" )
            for inter in interstitial_list:
                print(f"       {inter[0]}, {inter[1]} added")


        print(f"Mapping after structure manipulation           : {len(mapping)}/{len(stru_defect)}")

        ########################
        # change the IFCs
        ########################

        ifc_emb=np.zeros(shape=np.shape(ifc_pristine))

        if cut_off_mode=='auto':
            rc_1=100000 # very large value to include all the ifcs, no sparse matrix. 
            rc_2=0.99*min(np.array(structure_defect_wo_relax.lattice.abc)/2) # largest sphere inscribed in defect supercell, 
                                                                             # 0.99 to avoid problem with atoms at the border
        if cut_off_mode=='manual':
            rc_1=rc_1
            rc_2=rc_2

        # Set to zero if 2 atoms distance is bigger than rc_1

        for i,atom1 in enumerate(stru_emb):
            for j,atom2 in enumerate(stru_emb):
                dist_ij=np.sqrt(sum((atom1.coords-atom2.coords)**2))

                if dist_ij > rc_1:
                    ifc_emb[i][j]=np.zeros(shape=(3,3))

                else:
                    ifc_emb[i][j]=ifc_pristine[i][j]
                    
        # Set to doped phonons if 2 atoms are separated from defect by distance < R_c2
        print(f"\n Set IFC to explicit defect phonons calculations if both atoms are separated from defect by a distance < R_c2 = {round(rc_2,3)}")

        for i,atom1 in enumerate(stru_emb):
            for j,atom2 in enumerate(stru_emb):
                # structure centered around defect!!!, main defect is at [0,0,0]
                dist_1_from_defect=np.sqrt(sum((atom1.coords)**2))
                dist_2_from_defect=np.sqrt(sum((atom2.coords)**2))

                if dist_1_from_defect < rc_2 and dist_2_from_defect < rc_2:

                    if verbose > 0:
                        print(f"\n \n Atomic pair : {i,atom1} - {j,atom2}  \n")
                        print(f"atom1 : Dist. from. defect = {dist_1_from_defect}")
                        print(f"atom2 : Dist. from. defect = {dist_2_from_defect}")

                        print(f"Replacing pristine cell IFC = \n {ifc_pristine[i][j]}" )
                        print(f"by defect cell IFC = \n {ifc_defect[mapping.index(i)][mapping.index(j)]}" )
                        print(f"Diff IFC = \n {ifc_pristine[i][j]-ifc_defect[mapping.index(i)][mapping.index(j)]}" )


                    ifc_emb[i][j]=factor_ifc*ifc_defect[mapping.index(i)][mapping.index(j)]
                    
                # enforce ASR
        if asr==True:
            print(f"\n Enforce ASR")
            for alpha in [0,1,2]:
                for beta in [0,1,2]:
                    for i,atom1 in enumerate(stru_emb):
                        ifc_emb[i][i][alpha,beta] = -(accoustic_sum(ifc_emb,i)[alpha,beta]-ifc_emb[i][i][alpha,beta])
        
        ########################
        # change the nac params
        ########################
        
        if phonopy_pristine.nac_params is not None:
            
            # Simply copy the nac params from the pristine
            nac_params_emb=phonopy_pristine.nac_params.copy()

            if vacancies_list is not None:
                nac_params_emb["born"]=np.delete(nac_params_emb["born"],vacancies_list,0)
        else:
            nac_params_emb=None

        ########################

        print(f"\n Embedding procedure done")

        return cls(stru_pristine,stru_defect,stru_emb,ifc_emb,nac_params_emb)
    

    def get_gamma_freq_with_vec_abipy_fmt(self):
        """
        Compute with phonopy the Gamma phonons freq and vectors and returns them in abipy format
        such that ph_vec[iband][iatom] gives vector of iband,iatom
        
        Returns:
            phonons frequencies, phonon eigenvectors
        """

        ph_freq_phonopy,ph_vec_phonopy=self.get_frequencies_with_eigenvectors(q=[0,0,0])

        ph_freq = ph_freq_phonopy / (eV_to_THz)  # put it in eV
        ph_vec = ph_vec_phonopy.transpose().reshape(3*len(self.supercell),len(self.supercell),3)  
        
        return ph_freq, ph_vec
    
    
    def to_ddb(self,embedded_ddb_path='out_DDB',workdir=None):
        """
        Call the converter to go from phonopy to a DDB

        Args:
            embedded_ddb_path: filepath where to save the DDB 
            workdir: work directory for the conversion 
        """
        ddb_sc = phonopy_to_abinit(unit_cell=get_pmg_structure(self.supercell), supercell_matrix=[1,1,1], qpt_list=[[0,0,0]],
                                   out_ddb_path=embedded_ddb_path, force_constants=self.force_constants,
                                   born=self.nac_params, primitive_matrix=np.eye(3), symprec=1e-5,
                                   tolsym=None,workdir=workdir,nsym=1)

        return ddb_sc







#!/usr/bin/env python
r"""
GWR flow with convergence studies
=================================

This script shows how to compute the G0W0 corrections in silicon.
More specifically, we build a flow to analyze the convergence of the QP corrections
wrt to the number of bands in the self-energy. More complicated convergence studies
can be implemented on the basis of this example.
"""

import os
import sys
import abipy.data as data
import abipy.abilab as abilab

from abipy import flowtk


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_","flow_")

    #from abipy.flowtk.psrepos import get_repo_from_name
    #pseudos = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4").get_pseudos("stringent")

    scf_input = abilab.AbinitInput(structure=data.cif_file("si.cif"),
                                   pseudos=data.pseudos("14si.pspnc"))

    # Global variables.
    scf_input.set_vars(
        ecut=6,
        tolvrs=1e-8,
        nband=4,
        paral_kgb=1,
    )
    scf_input.set_kmesh(
        ngkpt=[2, 2, 2],
        shiftk=[0.0, 0.0, 0.0], # IMPORTANT: k-grid for GWR must be Gamma-centered.
    )

    flow = flowtk.Flow(workdir=options.workdir)

    # GS-SCF run to get the DEN, followed by direct diago to obtain green_nband bands.
    from abipy.flowtk.gwr_works import DirectDiagoWork, GWRSigmaConvWork
    green_nband = -1  # -1 this means full diago
    diago_work = DirectDiagoWork.from_scf_input(scf_input, green_nband)
    flow.register_work(diago_work)

    # Build template for GWR.
    gwr_template = scf_input.make_gwr_qprange_input(gwr_ntau=6, nband=8, ecuteps=4)

    # Two possibilities:
    # 1) Change the value of one variable:

    varname_values = ("nband", [8, 12, 14])

    # or take the Cartesian product of two or more variables with e.g.:
    #
    #varname_values = [
    #   ("gwr_ntau", [6, 8]),
    #   ("ecuteps", [2, 4]),
    #]

    gwr_work = GWRSigmaConvWork.from_varname_values(
            varname_values, gwr_template, den_node=diago_work[0], wfk_node=diago_work[1])
    flow.register_work(gwr_work)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""Work subclasses related to GS calculations."""
from __future__ import annotations

import json

from pymatgen.analysis.eos import EOS
from abipy.core.structure import Structure
from abipy.abio.inputs import AbinitInput
from abipy.electrons.gsr import GsrRobot
from .works import Work

__all__ = [
    "GsKmeshConvWork"
    "GsKmeshTsmearConvWork",
    "EosWork",
]


class GsKmeshConvWork(Work):
    """
    This work performs convergence studies of GS properties
    with respect to the k-mesh

    It produces ...

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GsKmeshConvWork
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput, nksmall_list: list) -> GsKmeshConvWork:
        """
        Build the work from a `scf_input` for a GS SCF run and a list
        with the smallest number of divisions for the k-mesh.
        """
        work = cls()
        for nksmall in nksmall_list:
            new_inp = scf_input.deepcopy()
            new_inp.set_autokmesh(nksmall)
            work.register_scf_task(new_inp)

        return work

    def on_all_ok(self): # pragma: no cover
        """
        This method is called when all tasks in the GsKmeshTsmearConvWork have reached S_OK.
        """
        with GsrRobot.from_work(self) as gsr_robot:
            df = gsr_robot.get_dataframe(with_geo=False)
            # Write excel file.
            basename = self.__class__.__name__
            df.to_excel(self.outdir.path_in(f"{basename}.xlsx"))

            with gsr_robot.get_pyscript(self.outdir.path_in("gsr_robot.py")) as script:
                script.add_text("""
# Quantities that should be tested for convergence.
abs_conv = {
"energy_per_atom": 1e-3,
"pressure": 1e-2,
"max_force": 1e-4,
}
items = abs_conv.keys()

robot.plot_convergence_items(items, sortby="nkpt", abs_conv=abs_conv)
""")

        return super().on_all_ok()



class GsKmeshTsmearConvWork(Work):
    """
    This work performs convergence studies of GS properties
    with respect to the k-mesh and the electronic smearing.

    It produces ...

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GsKmeshTsmearConvWork
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput, nksmall_list: list, tsmear_list: list) -> GsKmeshTsmearConvWork:
        """
        Build the work from a `scf_input` for a GS SCF run including `occopt`
        and a list with the smallest number of divisions for the k-mesh.
        """
        occopt = scf_input.get("occopt", default=None)
        if occopt is None or occopt <= 0:
            raise ValueError(f"scf_input should define occopt but found: {occopt}")

        work = cls()
        for tsmear in tsmear_list:
            for nksmall in nksmall_list:
                new_inp = scf_input.new_with_vars(tsmear=tsmear)
                new_inp.set_autokmesh(nksmall)
                work.register_scf_task(new_inp)

        return work

    def on_all_ok(self): # pragma: no cover
        """
        This method is called when all tasks in the GsKmeshTsmearConvWork have reached S_OK.
        """
        with GsrRobot.from_work(self) as gsr_robot:
            df = gsr_robot.get_dataframe(with_geo=False)
            # Write excel file.
            basename = self.__class__.__name__
            df.to_excel(self.outdir.path_in(f"{basename}.xlsx"))

            with gsr_robot.get_pyscript(self.outdir.path_in("gsr_robot.py")) as script:
                script.add_text("""
# Quantities that should be tested for convergence.
abs_conv = {
"energy_per_atom": 1e-3,
"pressure": 1e-2,
"max_force": 1e-4,
}
items = abs_conv.keys()

robot.plot_convergence_items(items, sortby="nkpt", hue="tsmear", abs_conv=abs_conv)
""")

        return super().on_all_ok()


class EosWork(Work):
    """
    Work to compute the Equation of State.
    The EOS is obtained by computing E(V) for several volumes around the input V0,
    The initial volumes are obtained by rescaling the input lattice vectors so that
    length proportions and angles are preserved.
    This guess is exact for cubic materials while other Bravais lattices require
    a constant-volume optimization of the cell geometry.

    If lattice_type=="cubic" and atomic positions are fixed by symmetry.
    use can use move_atoms=False to perform standard GS-SCF calculations.
    In all the other cases, E(V) is obtained by relaxing the atomic positions at fixed volume.

    The E(V) points are fitted at the end of the work and the results are saved in the
    `eos_data.json` file produced in the `outdata` directory.
    The file contains the energies, the volumes and the values of V0, B0, B1 obtained
    with different EOS models.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: EosWork
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput,
                       npoints=4, deltap_vol=0.25, ecutsm=0.5, move_atoms=True,
                       manager=None) -> EosWork:
        """
        Build an EosWork from an AbinitInput for GS-SCF.

        Args:
            scf_input: AbinitInput for GS-SCF used as template to generate the other inputs.
            npoints: Number of volumes generated on the right (left) of the equilibrium volume
                The total number of points is therefore 2 * n + 1.
            deltap_vol: Step of the linear mesh given in relative percentage of the equilibrium volume
                The step is thus: v0 * deltap_vol / 100.
            ecutsm: Value of ecutsm input variable. If `scf_input` does not provide ecutsm, this
                value will be used else the vale in `scf_input`.
            move_atoms: If True, a structural relaxation of ions is performed for each volume
                This is needed if the atomic positions are non fixed by symmetry.
            manager: TaskManager instance. Use default if None.
        """
        new_work = cls(manager=manager)

        structure = scf_input.structure
        lattice_type = structure.spget_lattice_type()
        assert lattice_type is not None

        dvol = structure.volume * deltap_vol / 100
        v0 = structure.volume - dvol * npoints
        new_work.input_volumes = [v0 + ipt * dvol for ipt in range(2 * npoints + 1)]

        if "ecutsm" not in scf_input:
            print("Input does not define ecutsm input variable.\n",
                  "A default value of %s will be added to all the EOS inputs" % ecutsm)

        for vol in new_work.input_volumes:
            # Build structure with new volume and generate new input.
            new_lattice = structure.lattice.scale(vol)
            new_structure = Structure(new_lattice, structure.species, structure.frac_coords)
            new_input = scf_input.new_with_structure(new_structure)

            # Add ecutsm if not already present.
            new_input.set_vars_ifnotin(ecutsm=ecutsm)

            if lattice_type == "cubic" and not move_atoms:
                # Perform GS calculations without moving atoms, cells do not need to be relaxed.
                new_input.pop_vars(["ionmov", "optcell", "ntime"])
                new_work.register_scf_task(new_input)
            else:
                # Constant-volume optimization of cell geometry + atoms.
                # (modify acell and rprim under constraint - normalize the vectors of rprim to generate the acell)
                # In principle one could take into account the symmetry of the lattice...
                new_input.set_vars_ifnotin(ionmov=2, ntime=50, optcell=3, dilatmx=1.05)
                new_work.register_relax_task(new_input)

        return new_work

    @classmethod
    def from_inputs(cls, inputs: list[AbinitInput], manager=None) -> EosWork:
        """
        Advanced interface to build an EosWork from an list of AbinitInputs.
        """
        new_work = cls(manager=manager)
        new_work.input_volumes = [inp.structure.lattice.volume for inp in inputs]
        for scf_inp in inputs:
            new_work.register_scf_task(scf_inp)

        return new_work

    def get_and_write_eosdata(self, write_json=True) -> dict:
        """
        Compute the EOS and produce a JSON file `eos_data.json` in outdata.
        """
        energies_ev, volumes = [], []
        for task in self:
            with task.open_gsr() as gsr:
                energies_ev.append(float(gsr.energy))
                volumes.append(float(gsr.structure.volume))

        eos_data = {"input_volumes_ang3": self.input_volumes,
                    "volumes_ang3": volumes,
                    "energies_ev": energies_ev}

        for model in EOS.MODELS:
            if model in ("deltafactor", "numerical_eos"): continue
            try:
                fit = EOS(model).fit(volumes, energies_ev)
                eos_data[model] = {k: float(v) for k, v in fit.results.items()}
            except Exception as exc:
                eos_data[model] = {"exception": str(exc)}

        if write_json:
            with open(self.outdir.path_in("eos_data.json"), "wt") as fh:
                json.dump(eos_data, fh, indent=4, sort_keys=True)

        return eos_data

    def on_all_ok(self): # pragma: no cover
        """
        This method is called when all tasks have reached S_OK.
        It reads the energies and the volumes from the GSR file, computes the EOS
        and produce a JSON file `eos_data.json` in outdata.
        """
        self.get_and_write_eosdata()
        return super().on_all_ok()
